import os 
from pathlib import Path
import yaml

_HERE = Path(__file__).resolve().parent

LOG_STORE_DIR = os.getenv('LOG_STORE_DIR', './runs')

def load_task_cfg(task): 

    # if task.yaml exists: 

    if os.path.exists(os.path.join(_HERE, "..", "cfgs", f"{task}.yaml")):

        with open(os.path.join(_HERE, "..", "cfgs", f"{task}.yaml"), "r") as f:
            return yaml.load(f, Loader = yaml.FullLoader)
        
    elif os.path.exists(os.path.join(_HERE, "..", "cfgs", f"{task}.py")):
        task_cfg = __import__(f"dart_physics.cfgs.{task}", fromlist = ["task_cfg"])

        return task_cfg.task_cfg
    
    else:
        print("First tried to load task from ", os.path.join(_HERE, "..", "cfgs", f"{task}.yaml"))
        print("Then tried to load task from ", os.path.join(_HERE, "..", "cfgs", f"{task}.py"))
        print("No task config found")
        raise ValueError(f"Task config {task} not found")
    
def load_robot_cfg(robot): 
    
    robot_cfg = __import__(f"dart_physics.robot_cfgs.{robot}", fromlist = ["robot_cfg"])

    return robot_cfg.robot_cfg

if __name__ == "__main__":

    cfg = load_robot_cfg("freefloating_dual_allegro")
    print(cfg)



from dart_physics.robot_cfgs import * 
from pathlib import Path

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos): 

    ctrl = np.zeros(32)
    ctrl[:16] = qpos[7:23]
    ctrl[16:] = qpos[30:46]
    return ctrl 

robot_cfg = {

    "name": "FreefloatingDualAllegro",

    "robots": { 
        "l_robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "wonik_allegro" / "left_hand.xml").as_posix(),
                    home_q = [-0.5, 0.0, 0.3, 0, 0, 0, 1] + [0.] * 16, 
                    freejoint = True, 
                ), 
        "r_robot": RobotConfig(
                    xml_path = (_HERE / "../assets" / "wonik_allegro" / "right_hand.xml").as_posix(),
                    home_q = [0.5, 0.0, 0.3, 0, 0, 0, 1] + [0.] * 16,
                    freejoint = True,
                ) 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": (7+16) * 2, 
        
    "points" : { 
        "left_wrist":  FingerKeypoints("left", 0, "left_wrist_target", "l_robot/palm", weld = True, \
                                       avp_transform = rot_z(90) @ rot_x(90), init_posquat = [-0.5, 0.0, 0.3, 0, 0, 0, 1]), 
        "right_wrist": FingerKeypoints("right", 0, "right_wrist_target", "r_robot/palm", weld = True, \
                                       avp_transform = rot_z(-90) @ rot_x(90), init_posquat = [0.5, 0.0, 0.3, 0, 0, 0, 1]),
        
        "left_ff":     FingerKeypoints("left", 9, "left_ff_tip_target", "l_robot/ff_tip", type = "site"),
        "left_mf":      FingerKeypoints("left", 14, "left_mf_tip_target", "l_robot/mf_tip", type = "site"),
        "left_rf":      FingerKeypoints("left", 19, "left_rf_tip_target", "l_robot/rf_tip", type = "site"),
        "left_th":      FingerKeypoints("left", 4, "left_th_tip_target", "l_robot/th_tip", type = "site"),
        
        "right_ff":     FingerKeypoints("right", 9, "right_ff_tip_target", "r_robot/ff_tip", type = "site"),
        "right_mf":     FingerKeypoints("right", 14, "right_mf_tip_target", "r_robot/mf_tip", type = "site"),
        "right_rf":     FingerKeypoints("right", 19, "right_rf_tip_target", "r_robot/rf_tip", type = "site"),
        "right_th":     FingerKeypoints("right", 4, "right_th_tip_target", "r_robot/th_tip", type = "site"),
    },

    "reverse_points": { 
        "left_wrist":  FingerKeypoints("left", 0, "left_wrist_target", "l_robot/palm", weld = True, \
                                       avp_transform = rot_z(90) @ rot_x(90), init_posquat = [-0.5, 0.0, 0.3, 0, 0, 0, 1]), 
        "right_wrist": FingerKeypoints("right", 0, "right_wrist_target", "r_robot/palm", weld = True, \
                                       avp_transform = rot_z(-90) @ rot_x(90), init_posquat = [0.5, 0.0, 0.3, 0, 0, 0, 1]),
        
        "left_ff":     FingerKeypoints("left", 9, "left_ff_tip_target", "l_robot/ff_tip", type = "site"),
        "left_mf":      FingerKeypoints("left", 14, "left_mf_tip_target", "l_robot/mf_tip", type = "site"),
        "left_rf":      FingerKeypoints("left", 19, "left_rf_tip_target", "l_robot/rf_tip", type = "site"),
        "left_th":      FingerKeypoints("left", 4, "left_th_tip_target", "l_robot/th_tip", type = "site"),
        
        "right_ff":     FingerKeypoints("right", 9, "right_ff_tip_target", "r_robot/ff_tip", type = "site"),
        "right_mf":     FingerKeypoints("right", 14, "right_mf_tip_target", "r_robot/mf_tip", type = "site"),
        "right_rf":     FingerKeypoints("right", 19, "right_rf_tip_target", "r_robot/rf_tip", type = "site"),
        "right_th":     FingerKeypoints("right", 4, "right_th_tip_target", "r_robot/th_tip", type = "site"),
    },

    "avp_calib": { 

        "left_ff": {"scale": 1.3, "offset": np.array([-0.02, 0.0, 0.0])},
        "left_mf": {"scale": 1.2, "offset": np.array([+0.00, 0.0, 0.0])},
        "left_rf": {"scale": 1.2, "offset": np.array([+0.02, 0.0, 0.0])},
        "left_th": {"scale": 1.0, "offset": np.array([+0.00, 0.0, 0.0])},

        "right_ff": {"scale": 1.3, "offset": np.array([0.0, -0.02, 0.0])},
        "right_mf": {"scale": 1.2, "offset": np.array([0.0, -0.00, 0.0])},
        "right_rf": {"scale": 1.2, "offset": np.array([0.0, +0.02, 0.0])},
        "right_th": {"scale": 1.0, "offset": np.array([0.0, +0.00, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(root = "left_wrist", target = "left_rf"), 
        IKTasks(root = "left_wrist", target = "left_mf"),
        IKTasks(root = "left_wrist", target = "left_ff"),
        IKTasks(root = "left_wrist", target = "left_th"),

        IKTasks(root = "right_wrist", target = "right_rf"),
        IKTasks(root = "right_wrist", target = "right_mf"),
        IKTasks(root = "right_wrist", target = "right_ff"),
        IKTasks(root = "right_wrist", target = "right_th"),
    ],

    "joints" : [ 
        # left hand 
        "l_robot/rfj0", "l_robot/rfj1", "l_robot/rfj2", "l_robot/rfj3",
        "l_robot/mfj0", "l_robot/mfj1", "l_robot/mfj2", "l_robot/mfj3",
        "l_robot/ffj0", "l_robot/ffj1", "l_robot/ffj2", "l_robot/ffj3",
        "l_robot/thj0", "l_robot/thj1", "l_robot/thj2", "l_robot/thj3",

        # right hand
        "r_robot/rfj0", "r_robot/rfj1", "r_robot/rfj2", "r_robot/rfj3",
        "r_robot/mfj0", "r_robot/mfj1", "r_robot/mfj2", "r_robot/mfj3",
        "r_robot/ffj0", "r_robot/ffj1", "r_robot/ffj2", "r_robot/ffj3",
        "r_robot/thj0", "r_robot/thj1", "r_robot/thj2", "r_robot/thj3",
    ],

    "bodies":  {
        "l_robot/palm": "left_allegro_palm", 
        "r_robot/palm": "right_allegro_palm",


        # left ff 
        "l_robot/ff_base":   "left_allegro_ff_base",   "l_robot/ff_distal":   "left_allegro_ff_distal", 
        "l_robot/ff_medial": "left_allegro_ff_medial", "l_robot/ff_proximal": "left_allegro_ff_proximal", 
        "l_robot/ff_tip": "left_allegro_ff_tip", 

        # left mf 
        "l_robot/mf_base":   "left_allegro_mf_base",   "l_robot/mf_distal":   "left_allegro_mf_distal", 
        "l_robot/mf_medial": "left_allegro_mf_medial", "l_robot/mf_proximal": "left_allegro_mf_proximal", 
        "l_robot/mf_tip": "left_allegro_mf_tip", 

        # left rf 
        "l_robot/rf_base":   "left_allegro_rf_base",   "l_robot/rf_distal":   "left_allegro_rf_distal",
        "l_robot/rf_medial": "left_allegro_rf_medial", "l_robot/rf_proximal": "left_allegro_rf_proximal",
        "l_robot/rf_tip": "left_allegro_rf_tip",

        # left th
        "l_robot/th_base":   "left_allegro_th_base",   "l_robot/th_distal":   "left_allegro_th_distal",
        "l_robot/th_medial": "left_allegro_th_medial", "l_robot/th_proximal": "left_allegro_th_proximal",
        "l_robot/th_tip": "left_allegro_th_tip",

        # right ff
        "r_robot/ff_base":   "right_allegro_ff_base",   "r_robot/ff_distal":   "right_allegro_ff_distal",
        "r_robot/ff_medial": "right_allegro_ff_medial", "r_robot/ff_proximal": "right_allegro_ff_proximal",
        "r_robot/ff_tip": "right_allegro_ff_tip",

        # right mf
        "r_robot/mf_base":   "right_allegro_mf_base",   "r_robot/mf_distal":   "right_allegro_mf_distal",
        "r_robot/mf_medial": "right_allegro_mf_medial", "r_robot/mf_proximal": "right_allegro_mf_proximal",
        "r_robot/mf_tip": "right_allegro_mf_tip",

        # right rf
        "r_robot/rf_base":   "right_allegro_rf_base",   "r_robot/rf_distal":   "right_allegro_rf_distal",
        "r_robot/rf_medial": "right_allegro_rf_medial", "r_robot/rf_proximal": "right_allegro_rf_proximal",
        "r_robot/rf_tip": "right_allegro_rf_tip",

        # right th
        "r_robot/th_base":   "right_allegro_th_base",   "r_robot/th_distal":   "right_allegro_th_distal",
        "r_robot/th_medial": "right_allegro_th_medial", "r_robot/th_proximal": "right_allegro_th_proximal",
        "r_robot/th_tip": "right_allegro_th_tip",

    }, 

    "usdz_geoms":  [ 

        # LEFT 
        BodyInfo(
            name = "left_allegro_ff_base", 
            geoms = [GeomInfo(mesh = "link_0.0")], 
        ), 
        BodyInfo(
            name = "left_allegro_ff_distal", 
            geoms = [GeomInfo(mesh = "link_3.0")], 
        ),
        BodyInfo(
            name = "left_allegro_ff_medial", 
            geoms = [GeomInfo(mesh = "link_2.0")], 
        ),
        BodyInfo(
            name = "left_allegro_ff_proximal", 
            geoms = [GeomInfo(mesh = "link_1.0")], 
        ),
        BodyInfo(
            name = "left_allegro_ff_tip", 
            geoms = [GeomInfo(mesh = "link_3.0_tip", pos=[0, 0, 0.0267])], 
        ),
        BodyInfo(
            name = "left_allegro_mf_base", 
            geoms = [GeomInfo(mesh = "link_0.0")], 
        ),
        BodyInfo(
            name = "left_allegro_mf_distal", 
            geoms = [GeomInfo(mesh = "link_3.0")], 
        ),
        BodyInfo(
            name = "left_allegro_mf_medial", 
            geoms = [GeomInfo(mesh = "link_2.0")], 
        ),
        BodyInfo(
            name = "left_allegro_mf_proximal", 
            geoms = [GeomInfo(mesh = "link_1.0")], 
        ),
        BodyInfo(
            name = "left_allegro_mf_tip", 
            geoms = [GeomInfo(mesh = "link_3.0_tip", pos=[0, 0, 0.0267])], 
        ),
        BodyInfo(
            name = "left_allegro_rf_base", 
            geoms = [GeomInfo(mesh = "link_0.0")], 
        ),
        BodyInfo(
            name = "left_allegro_rf_distal", 
            geoms = [GeomInfo(mesh = "link_3.0")], 
        ),
        BodyInfo(
            name = "left_allegro_rf_medial", 
            geoms = [GeomInfo(mesh = "link_2.0")], 
        ),
        BodyInfo(
            name = "left_allegro_rf_proximal", 
            geoms = [GeomInfo(mesh = "link_1.0")], 
        ),
        BodyInfo(
            name = "left_allegro_rf_tip", 
            geoms = [GeomInfo(mesh = "link_3.0_tip", pos=[0, 0, 0.0267])], 
        ),
        BodyInfo(
            name = "left_allegro_th_base", 
            geoms = [GeomInfo(mesh = "link_12.0_left", quat = [0, 1, 0, 0])], 
        ),
        BodyInfo(
            name = "left_allegro_th_distal", 
            geoms = [GeomInfo(mesh = "link_15.0")], 
        ),
        BodyInfo( 
            name = "left_allegro_th_medial",
            geoms = [GeomInfo(mesh = "link_14.0")],
        ), 
        BodyInfo(
            name = "left_allegro_th_proximal",
            geoms = [GeomInfo(mesh = "link_13.0")],
        ),
        BodyInfo(
            name = "left_allegro_th_tip",
            geoms = [GeomInfo(mesh = "link_15.0_tip", pos=[0, 0, 0.0423])],
        ),
        BodyInfo(
            name = "left_allegro_palm", 
            geoms = [GeomInfo(mesh = "base_link_left", quat = [1, -1, 0, 0])], 
        ),



        # LEFT 

        BodyInfo(
            name = "right_allegro_ff_base", 
            geoms = [GeomInfo(mesh = "link_0.0")], 
        ),
        BodyInfo(
            name = "right_allegro_ff_distal", 
            geoms = [GeomInfo(mesh = "link_3.0")], 
        ),
        BodyInfo(
            name = "right_allegro_ff_medial", 
            geoms = [GeomInfo(mesh = "link_2.0")], 
        ),
        BodyInfo(
            name = "right_allegro_ff_proximal", 
            geoms = [GeomInfo(mesh = "link_1.0")], 
        ),
        BodyInfo(
            name = "right_allegro_ff_tip", 
            geoms = [GeomInfo(mesh = "link_3.0_tip", pos=[0, 0, 0.0267])], 
        ),
        BodyInfo(
            name = "right_allegro_mf_base", 
            geoms = [GeomInfo(mesh = "link_0.0")], 
        ),
        BodyInfo(
            name = "right_allegro_mf_distal", 
            geoms = [GeomInfo(mesh = "link_3.0")], 
        ),
        BodyInfo(
            name = "right_allegro_mf_medial", 
            geoms = [GeomInfo(mesh = "link_2.0")], 
        ),
        BodyInfo(
            name = "right_allegro_mf_proximal", 
            geoms = [GeomInfo(mesh = "link_1.0")], 
        ),
        BodyInfo(
            name = "right_allegro_mf_tip", 
            geoms = [GeomInfo(mesh = "link_3.0_tip", pos=[0, 0, 0.0267])], 
        ),
        BodyInfo(
            name = "right_allegro_rf_base", 
            geoms = [GeomInfo(mesh = "link_0.0")], 
        ),
        BodyInfo(
            name = "right_allegro_rf_distal", 
            geoms = [GeomInfo(mesh = "link_3.0")], 
        ),
        BodyInfo(
            name = "right_allegro_rf_medial", 
            geoms = [GeomInfo(mesh = "link_2.0")], 
        ),
        BodyInfo(
            name = "right_allegro_rf_proximal", 
            geoms = [GeomInfo(mesh = "link_1.0")], 
        ),
        BodyInfo(
            name = "right_allegro_rf_tip", 
            geoms = [GeomInfo(mesh = "link_3.0_tip", pos=[0, 0, 0.0267])], 
        ),
        BodyInfo(
            name = "right_allegro_th_base", 
            geoms = [GeomInfo(mesh = "link_12.0_right")], 
        ),
        BodyInfo(
            name = "right_allegro_th_distal", 
            geoms = [GeomInfo(mesh = "link_15.0")], 
        ),
        BodyInfo( 
            name = "right_allegro_th_medial",
            geoms = [GeomInfo(mesh = "link_14.0")],
        ),
        BodyInfo(
            name = "right_allegro_th_proximal",
            geoms = [GeomInfo(mesh = "link_13.0")],
        ),
        BodyInfo(
            name = "right_allegro_th_tip",
            geoms = [GeomInfo(mesh = "link_15.0_tip", pos=[0, 0, 0.0423])],
        ),
        BodyInfo(
            name = "right_allegro_palm", 
            geoms = [GeomInfo(mesh = "base_link")], 
        ),

    ]

}


if __name__ == "__main__": 


    import trimesh 
    from scipy.spatial.transform import Rotation as R
    import os 

    asset_root = os.path.join(_HERE, "../assets", "wonik_allegro", "assets")
    usdz_root = os.path.join(_HERE, "../assets", "wonik_allegro", "transformed_stl")\

    os.makedirs(usdz_root, exist_ok = True)

    input_list = [f"{k}" for k in robot_cfg["bodies"].values()]
    output_string = str(input_list).replace("'", '"')
    print(output_string)

    for body in robot_cfg["usdz_geoms"]:

        body: BodyInfo

        meshes = [] 
        for i, geom in enumerate(body.geoms): 
            geom: GeomInfo
            mesh = trimesh.load_mesh(f"{asset_root}/{geom.mesh}.stl")
            pos = np.array(geom.pos)
            quat = np.array(geom.quat)

            mat = np.eye(4)
            mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
            mat[:3, 3] = pos

            mesh.visual.vertex_colors = np.array([0, 0, 0, 255], dtype = np.uint8)

            mesh.apply_transform(mat)

            meshes.append(mesh)
        
        meshes = trimesh.util.concatenate(meshes)
        # mat = np.eye(4)
        # pos = np.array(body.pos)
        # quat = np.array(body.quat)

        # mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
        # mat[:3, 3] = pos

        # meshes.apply_transform(mat)

        meshes.export(f"{usdz_root}/{body.name}.glb")





from dart_physics.robot_cfgs import * 
from pathlib import Path
import dart_physics.mink as mink 

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos_target): 

    ctrl = np.zeros(14)
    ctrl[:6] = qpos_target[0:6]
    # ctrl[7:13] = qpos_target[8:14]

    # left gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "left_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "left_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[8] - 0.01

    # if the fingers are close  (dist 0), it should command 0.788 
    # if the fingers are far away  (dist 0.1), it should command 0 
    # otherwise, it should be linearly interpolated between 0.788 and 0 
    ctrl[6] = 0.788 - 0.788 * finger_dist / 0.1 
    
    ctrl[7:13] = qpos_target[14:20]

    # right gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "right_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "right_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[16] + 0.01

    ctrl[13] = 0.788 - 0.788 * finger_dist / 0.1 

    return ctrl 

joints = ["shoulder_pan", "shoulder_lift", "forearm_link", "wrist_1", "wrist_2", "wrist_3", "right_driver_joint", "right_coupler_joint",
          "right_spring_link_joint", "right_follower_joint", "left_driver_joint", "left_coupler_joint",
          "left_spring_link_joint", "left_follower_joint"]

gripper_bodies = ["2f85_base", "base_mount", "left_coupler", "left_driver", "left_follower", "left_pad", 
          "left_silicone_pad", "left_spring_link", "right_coupler", "right_driver", "right_follower", "right_pad", "right_silicone_pad", "right_spring_link"]

arm_bodies = ["base", "shoulder_link", "upper_arm_link", "forearm_link", "wrist_1_link", "wrist_2_link", "wrist_3_link"]

robot_cfg = {

    "name": "DualUR5e2F85",

    "robots": { 
        "l_robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "ur5e_2f85" / "ur5e.xml").as_posix(),
                    home_q = [-1.5708, -1.5708, 1.5708, -1.5708, -1.5708, 0, 0, 0, 0, 0, 0, 0, 0, 0], 
                    freejoint = False, 
                    attach_to = [-0.5, -0.0, 0.0, 1, 0, 0, 0 ], 
                    parallel_jaw = True, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "left_silicone_pad", \
                                      body_name = "leftarm_left_finger_tip", pos = [0.0, 0.0, 0.02]),
                        NewBodyConfig(attach_to = "left_follower", \
                                        body_name = "leftarm_left_finger_base", pos = [0, 0, 0.0]), 
                        # NewBodyConfig(attach_to = "left_right_finger_link", \
                        #                 body_name = "leftarm_right_finger_tip", pos = [0.0, 0.05, 0.02]),
                        # NewBodyConfig(attach_to = "left_left_finger_link", \
                        #                 body_name = "leftarm_left_finger_tip", pos = [0, -0.05, 0.02]),

                        NewBodyConfig(attach_to = "right_silicone_pad", \
                                        body_name = "leftarm_right_finger_tip", pos = [0.0, 0.0, 0.02]),
                        NewBodyConfig(attach_to = "right_follower", \
                                        body_name = "leftarm_right_finger_base", pos = [0, 0, 0.0]),
                        # NewBodyConfig(attach_to = "right_right_finger_link", \
                        #                 body_name = "rightarm_right_finger_tip", pos = [0.0, 0.05, 0.02]),
                        # NewBodyConfig(attach_to = "right_left_finger_link", \
                        #                 body_name = "rightarm_left_finger_tip", pos = [0, -0.05, 0.02]),
                    ],
                ), 

        "r_robot": RobotConfig(
                    xml_path = (_HERE / "../assets" / "ur5e_2f85" / "ur5e.xml").as_posix(),
                    home_q = [-1.5708, -1.5708, 1.5708, -1.5708, -1.5708, 0, 0, 0, 0, 0, 0, 0, 0, 0], 
                    freejoint = False, 
                    attach_to = [0.5, -0.0, 0.0, 1, 0, 0, 0 ], 
                    parallel_jaw = True, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "left_silicone_pad", \
                                      body_name = "rightarm_left_finger_tip", pos = [0.0, 0.0, 0.02]),
                        NewBodyConfig(attach_to = "left_follower", \
                                      body_name = "rightarm_left_finger_base", pos = [0.0, 0.0, 0.0]),
                        # NewBodyConfig(attach_to = "left_right_finger_link", \
                        #                 body_name = "leftarm_right_finger_tip", pos = [0.0, 0.05, 0.02]),
                        # NewBodyConfig(attach_to = "left_left_finger_link", \
                        #                 body_name = "leftarm_left_finger_tip", pos = [0, -0.05, 0.02]),

                        NewBodyConfig(attach_to = "right_silicone_pad", \
                                        body_name = "rightarm_right_finger_tip", pos = [0.0, 0.0, 0.02]),
                        NewBodyConfig(attach_to = "right_follower", \
                                        body_name = "rightarm_right_finger_base", pos = [0, 0, 0.0]),
                        # NewBodyConfig(attach_to = "right_right_finger_link", \
                        #                 body_name = "rightarm_right_finger_tip", pos = [0.0, 0.05, 0.02]),
                        # NewBodyConfig(attach_to = "right_left_finger_link", \
                        #                 body_name = "rightarm_left_finger_tip", pos = [0, -0.05, 0.02]),
                    ],
                ) 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": 28, 
        
    "points" : { 

        "left_lb":     FingerKeypoints("left", 3, "left_lb_target", "l_robot/leftarm_left_finger_base", type = "body"),
        "left_lf":      FingerKeypoints("left", 4, "left_lf_target", "l_robot/leftarm_left_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("left", 8, "left_rb_target", "l_robot/leftarm_right_finger_base", type = "body"),
        "left_rf":      FingerKeypoints("left", 9, "left_rf_target", "l_robot/leftarm_right_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("right", 3, "right_lb_target", "r_robot/rightarm_left_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("right", 4, "right_lf_target", "r_robot/rightarm_left_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("right", 8, "right_rb_target", "r_robot/rightarm_right_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("right", 9, "right_rf_target", "r_robot/rightarm_right_finger_tip", type = "body"),

    },

    "reverse_points": { 

        "left_lb":     FingerKeypoints("left", 3, "left_lb_target", "l_robot/leftarm_left_finger_base", type = "body"),
        "left_lf":      FingerKeypoints("left", 4, "left_lf_target", "l_robot/leftarm_left_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("left", 8, "left_rb_target", "l_robot/leftarm_right_finger_base", type = "body"),
        "left_rf":      FingerKeypoints("left", 9, "left_rf_target", "l_robot/leftarm_right_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("right", 3, "right_lb_target", "r_robot/rightarm_left_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("right", 4, "right_lf_target", "r_robot/rightarm_left_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("right", 8, "right_rb_target", "r_robot/rightarm_right_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("right", 9, "right_rf_target", "r_robot/rightarm_right_finger_tip", type = "body"),
    },

    "avp_calib": { 

        "left_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

        "right_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(target = "left_lb"), 
        IKTasks(target = "left_lf"),
        IKTasks(target = "left_rb"),
        IKTasks(target = "left_rf"),

        IKTasks(target = "right_lb"),
        IKTasks(target = "right_lf"),
        IKTasks(target = "right_rb"),
        IKTasks(target = "right_rf"),
    ],

    "joints" : [f'l_robot/{j}' for j in joints] + [f'r_robot/{j}' for j in joints],

    "bodies":  {},

} 

# add the right arm bodies 
robot_cfg["bodies"].update({f"r_robot/{b}": f"2f85_{b}_right" for b in gripper_bodies})
robot_cfg["bodies"].update({f"r_robot/{b}": f"ur5e_{b}_right" for b in arm_bodies})
robot_cfg["bodies"].update({f"l_robot/{b}": f"2f85_{b}_left" for b in gripper_bodies})
robot_cfg["bodies"].update({f"l_robot/{b}": f"ur5e_{b}_left" for b in arm_bodies})



if __name__ == "__main__": 

    import trimesh 
    from scipy.spatial.transform import Rotation as R
    import os 

    asset_root = os.path.join(_HERE, "../assets", "ur5e_2f85", "assets")
    usdz_root = os.path.join(_HERE, "../assets", "ur5e_2f85", "transformed_stl")\

    os.makedirs(usdz_root, exist_ok = True)

    print([f"{k}" for k in robot_cfg["bodies"].values()])

    for k, v in robot_cfg["usdz_geoms"].items():

        meshes = [] 
        for i, geom in enumerate(v): 
            geom: GeomInfo
            mesh = trimesh.load_mesh(f"{asset_root}/{geom.mesh}.stl")
            pos = np.array(geom.pos)
            quat = np.array(geom.quat)

            mesh.visual.vertex_colors = np.array([0, 0, 0, 255], dtype = np.uint8)

            mat = np.eye(4)
            mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
            mat[:3, 3] = pos

            mesh.apply_transform(mat)

            meshes.append(mesh)

        scale = [mesh.scale for mesh in meshes]
        avg_scale = np.mean(scale, axis = 0)
        print(k, avg_scale)
        
        meshes = trimesh.util.concatenate(meshes)

        if avg_scale > 100: 

            meshes.apply_scale(0.001)

        if avg_scale < 1: 

            meshes.apply_scale(1)
        
        meshes.export(f"{usdz_root}/{k}.glb")





"""Frame task implementation."""

from __future__ import annotations

from typing import Optional

import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from ..lie import SE3
from .exceptions import TargetNotSet, TaskDefinitionError
from .task import Task


class FrameTask(Task):
    """Regulate the pose of a frame expressed in the world frame.

    Attributes:
        frame_name: Name of the frame to regulate, typically the name of body, geom
            or site in the robot model.
        frame_type: The frame type: `body`, `geom` or `site`.
        transform_frame_to_world: Target pose of the frame.
    """

    k: int = 6
    transform_target_to_world: Optional[SE3]

    def __init__(
        self,
        frame_name: str,
        frame_type: str,
        position_cost: npt.ArrayLike,
        orientation_cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(cost=np.zeros((self.k,)), gain=gain, lm_damping=lm_damping)
        self.frame_name = frame_name
        self.frame_type = frame_type
        self.position_cost = position_cost
        self.orientation_cost = orientation_cost
        self.transform_target_to_world = None

        self.set_position_cost(position_cost)
        self.set_orientation_cost(orientation_cost)

    def set_position_cost(self, position_cost: npt.ArrayLike) -> None:
        position_cost = np.atleast_1d(position_cost)
        if position_cost.ndim != 1 or position_cost.shape[0] not in (1, 3):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} position cost should be a vector of shape "
                "1 (aka identical cost for all coordinates) or (3,) but got "
                f"{position_cost.shape}"
            )
        if not np.all(position_cost >= 0.0):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} position cost should be >= 0"
            )
        self.cost[:3] = position_cost

    def set_orientation_cost(self, orientation_cost: npt.ArrayLike) -> None:
        orientation_cost = np.atleast_1d(orientation_cost)
        if orientation_cost.ndim != 1 or orientation_cost.shape[0] not in (1, 3):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} orientation cost should be a vector of "
                "shape 1 (aka identical cost for all coordinates) or (3,) but got "
                f"{orientation_cost.shape}"
            )
        if not np.all(orientation_cost >= 0.0):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} position cost should be >= 0"
            )
        self.cost[3:] = orientation_cost

    def set_target(self, transform_target_to_world: SE3) -> None:
        """Set the target pose.

        Args:
            transform_target_to_world: Transform from the task target frame to the
                world frame.
        """
        self.transform_target_to_world = transform_target_to_world.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target pose from a given robot configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        self.set_target(
            configuration.get_transform_frame_to_world(self.frame_name, self.frame_type)
        )

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the frame task error.

        This error is a twist :math:`e(q) \in se(3)` expressed in the local frame,
        i.e., it is a body twist. It is computed by taking the right-minus difference
        between the target pose :math:`T_{0t}` and current frame pose :math:`T_{0b}`:

        .. math::

            e(q) := {}_b \xi_{0b} = -(T_{t0} \ominus T_{b0})
            = -\log(T_{t0} \cdot T_{0b}) = -\log(T_{tb}) = \log(T_{bt})

        where :math:`b` denotes our frame, :math:`t` the target frame and
        :math:`0` the inertial frame.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Frame task error vector :math:`e(q)`.
        """
        if self.transform_target_to_world is None:
            raise TargetNotSet(self.__class__.__name__)

        transform_frame_to_world = configuration.get_transform_frame_to_world(
            self.frame_name, self.frame_type
        )
        return self.transform_target_to_world.minus(transform_frame_to_world)

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        r"""Compute the frame task Jacobian.

        The derivation of the formula for this Jacobian is detailed in
        [FrameTaskJacobian]_.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Frame task jacobian :math:`J(q)`.
        """
        if self.transform_target_to_world is None:
            raise TargetNotSet(self.__class__.__name__)

        jac = configuration.get_frame_jacobian(self.frame_name, self.frame_type)

        transform_frame_to_world = configuration.get_transform_frame_to_world(
            self.frame_name, self.frame_type
        )

        T_tb = self.transform_target_to_world.inverse() @ transform_frame_to_world
        return -T_tb.jlog() @ jac



"""Relative frame task implementation."""

from __future__ import annotations

from typing import Optional

import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from ..lie import SE3
from .exceptions import TargetNotSet, TaskDefinitionError
from .task import Task


class RelativeFrameTask(Task):
    """Regulate the pose of a frame relative to another frame.

    Attributes:
        frame_name: Name of the frame to regulate, typically the name of body, geom
            or site in the robot model.
        frame_type: The frame type: `body`, `geom` or `site`.
        root_name: Name of the frame the task is relative to.
        root_type: The root frame type: `body`, `geom` or `site`.
        transform_target_to_root: Target pose in the root frame.
    """

    k: int = 6
    transform_target_to_root: Optional[SE3]

    def __init__(
        self,
        frame_name: str,
        frame_type: str,
        root_name: str,
        root_type: str,
        position_cost: npt.ArrayLike,
        orientation_cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(cost=np.zeros((self.k,)), gain=gain, lm_damping=lm_damping)
        self.frame_name = frame_name
        self.frame_type = frame_type
        self.root_name = root_name
        self.root_type = root_type
        self.position_cost = position_cost
        self.orientation_cost = orientation_cost
        self.transform_target_to_root = None

        self.set_position_cost(position_cost)
        self.set_orientation_cost(orientation_cost)
        
    def set_position_cost(self, position_cost: npt.ArrayLike) -> None:
        position_cost = np.atleast_1d(position_cost)
        if position_cost.ndim != 1 or position_cost.shape[0] not in (1, 3):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} position cost should be a vector of shape "
                "1 (aka identical cost for all coordinates) or (3,) but got "
                f"{position_cost.shape}"
            )
        if not np.all(position_cost >= 0.0):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} position cost should be >= 0"
            )
        self.cost[:3] = position_cost

    def set_orientation_cost(self, orientation_cost: npt.ArrayLike) -> None:
        orientation_cost = np.atleast_1d(orientation_cost)
        if orientation_cost.ndim != 1 or orientation_cost.shape[0] not in (1, 3):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} orientation cost should be a vector of "
                "shape 1 (aka identical cost for all coordinates) or (3,) but got "
                f"{orientation_cost.shape}"
            )
        if not np.all(orientation_cost >= 0.0):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} position cost should be >= 0"
            )
        self.cost[3:] = orientation_cost

    def set_target(self, transform_target_to_root: SE3) -> None:
        """Set the target pose in the root frame.

        Args:
            transform_target_to_root: Transform from the task target frame to the
                root frame.
        """
        self.transform_target_to_root = transform_target_to_root.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target pose from a given robot configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        self.set_target(
            configuration.get_transform(
                self.frame_name,
                self.frame_type,
                self.root_name,
                self.root_type,
            )
        )

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        if self.transform_target_to_root is None:
            raise TargetNotSet(self.__class__.__name__)

        transform_frame_to_root = configuration.get_transform(
            self.frame_name,
            self.frame_type,
            self.root_name,
            self.root_type,
        )
        return transform_frame_to_root.rminus(self.transform_target_to_root)

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        if self.transform_target_to_root is None:
            raise TargetNotSet(self.__class__.__name__)

        jacobian_frame_in_frame = configuration.get_frame_jacobian(
            self.frame_name, self.frame_type
        )
        jacobian_root_in_root = configuration.get_frame_jacobian(
            self.root_name, self.root_type
        )

        transform_frame_to_root = configuration.get_transform(
            self.frame_name,
            self.frame_type,
            self.root_name,
            self.root_type,
        )
        transform_frame_to_target = (
            self.transform_target_to_root.inverse() @ transform_frame_to_root
        )

        return transform_frame_to_target.jlog() @ (
            jacobian_frame_in_frame
            - transform_frame_to_root.inverse().adjoint() @ jacobian_root_in_root
        )



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import dart_physics.avp_stream3.grpc_msg.dexterity_pb2 as dexterity__pb2


class HandTrackingServiceStub(object):
    """The hand tracking service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StreamHandUpdatesandGetSimStates = channel.stream_stream(
                '/dexterity.HandTrackingService/StreamHandUpdatesandGetSimStates',
                request_serializer=dexterity__pb2.HandUpdate.SerializeToString,
                response_deserializer=dexterity__pb2.SimStates.FromString,
                )
        self.StreamHandUpdatesandGetStereoImages = channel.stream_stream(
                '/dexterity.HandTrackingService/StreamHandUpdatesandGetStereoImages',
                request_serializer=dexterity__pb2.HandUpdate.SerializeToString,
                response_deserializer=dexterity__pb2.StereoImage.FromString,
                )
        self.Check = channel.unary_unary(
                '/dexterity.HandTrackingService/Check',
                request_serializer=dexterity__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=dexterity__pb2.HealthCheckResponse.FromString,
                )


class HandTrackingServiceServicer(object):
    """The hand tracking service definition.
    """

    def StreamHandUpdatesandGetSimStates(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamHandUpdatesandGetStereoImages(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Check(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HandTrackingServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StreamHandUpdatesandGetSimStates': grpc.stream_stream_rpc_method_handler(
                    servicer.StreamHandUpdatesandGetSimStates,
                    request_deserializer=dexterity__pb2.HandUpdate.FromString,
                    response_serializer=dexterity__pb2.SimStates.SerializeToString,
            ),
            'StreamHandUpdatesandGetStereoImages': grpc.stream_stream_rpc_method_handler(
                    servicer.StreamHandUpdatesandGetStereoImages,
                    request_deserializer=dexterity__pb2.HandUpdate.FromString,
                    response_serializer=dexterity__pb2.StereoImage.SerializeToString,
            ),
            'Check': grpc.unary_unary_rpc_method_handler(
                    servicer.Check,
                    request_deserializer=dexterity__pb2.HealthCheckRequest.FromString,
                    response_serializer=dexterity__pb2.HealthCheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'dexterity.HandTrackingService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class HandTrackingService(object):
    """The hand tracking service definition.
    """

    @staticmethod
    def StreamHandUpdatesandGetSimStates(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/dexterity.HandTrackingService/StreamHandUpdatesandGetSimStates',
            dexterity__pb2.HandUpdate.SerializeToString,
            dexterity__pb2.SimStates.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StreamHandUpdatesandGetStereoImages(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/dexterity.HandTrackingService/StreamHandUpdatesandGetStereoImages',
            dexterity__pb2.HandUpdate.SerializeToString,
            dexterity__pb2.StereoImage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Check(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/dexterity.HandTrackingService/Check',
            dexterity__pb2.HealthCheckRequest.SerializeToString,
            dexterity__pb2.HealthCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""All kinematic tasks derive from the :class:`Task` base class."""

import abc
from typing import NamedTuple

import numpy as np

from ..configuration import Configuration
from .exceptions import InvalidDamping, InvalidGain


class Objective(NamedTuple):
    r"""Quadratic objective of the form :math:`\frac{1}{2} x^T H x + c^T x`."""

    H: np.ndarray
    """Hessian matrix, of shape (n_v, n_v)"""
    c: np.ndarray
    """Linear vector, of shape (n_v,)."""

    def value(self, x: np.ndarray) -> float:
        """Returns the value of the objective at the input vector."""
        return x.T @ self.H @ x + self.c @ x


class Task(abc.ABC):
    r"""Abstract base class for kinematic tasks.

    Subclasses must implement the configuration-dependent task error
    :py:meth:`~Task.compute_error` and Jacobian :py:meth:`~Task.compute_jacobian`
    functions.

    The error function :math:`e(q) \in \mathbb{R}^{k}` is the quantity that
    the task aims to drive to zero (:math:`k` is the dimension of the
    task). It appears in the first-order task dynamics:

    .. math::

        J(q) \Delta q = -\alpha e(q)

    The Jacobian matrix :math:`J(q) \in \mathbb{R}^{k \times n_v}`, with
    :math:`n_v` the dimension of the robot's tangent space, is the
    derivative of the task error :math:`e(q)` with respect to the
    configuration :math:`q \in \mathbb{R}^{n_q}`. The configuration displacement
    :math:`\Delta q` is the output of inverse kinematics; we divide it by dt to get a
    commanded velocity.

    In the first-order task dynamics, the error :math:`e(q)` is multiplied
    by the task gain :math:`\alpha \in [0, 1]`. This gain can be 1.0 for
    dead-beat control (*i.e.* converge as fast as possible), but might be
    unstable as it neglects our first-order approximation. Lower values
    cause slow down the task, similar to low-pass filtering.
    """

    def __init__(
        self,
        cost: np.ndarray,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        """Constructor.

        Args:
            cost: Cost vector with the same dimension as the error of the task.
            gain: Task gain alpha in [0, 1] for additional low-pass filtering. Defaults
                to 1.0 (no filtering) for dead-beat control.
            lm_damping: Unitless scale of the Levenberg-Marquardt (only when the error
            is large) regularization term, which helps when targets are infeasible.
            Increase this value if the task is too jerky under unfeasible targets, but
            beware that a larger damping slows down the task.
        """
        if not 0.0 <= gain <= 1.0:
            raise InvalidGain("`gain` must be in the range [0, 1]")

        if lm_damping < 0.0:
            raise InvalidDamping("`lm_damping` must be >= 0")

        self.cost = cost
        self.gain = gain
        self.lm_damping = lm_damping

    @abc.abstractmethod
    def compute_error(self, configuration: Configuration) -> np.ndarray:
        """Compute the task error at the current configuration.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task error vector :math:`e(q)`.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        """Compute the task Jacobian at the current configuration.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Task jacobian :math:`J(q)`.
        """
        raise NotImplementedError

    def compute_qp_objective(self, configuration: Configuration) -> Objective:
        r"""Compute the matrix-vector pair :math:`(H, c)` of the QP objective.

        This pair is such that the contribution of the task to the QP objective is:

        .. math::

            \| J \Delta q + \alpha e \|_{W}^2 = \frac{1}{2} \Delta q^T H
            \Delta q + c^T q

        The weight matrix :math:`W \in \mathbb{R}^{k \times k}` weights and
        normalizes task coordinates to the same unit. The unit of the overall
        contribution is [cost]^2.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Pair :math:`(H(q), c(q))`.
        """
        jacobian = self.compute_jacobian(configuration)  # (k, nv)
        minus_gain_error = -self.gain * self.compute_error(configuration)  # (k,)

        weight = np.diag(self.cost)
        weighted_jacobian = weight @ jacobian
        weighted_error = weight @ minus_gain_error

        mu = self.lm_damping * weighted_error @ weighted_error
        eye_tg = np.eye(configuration.model.nv)

        H = weighted_jacobian.T @ weighted_jacobian + mu * eye_tg  # (nv, nv)
        c = -weighted_error.T @ weighted_jacobian  # (nv,)

        return Objective(H, c)



"""Joint position limit."""

import mujoco
import numpy as np

from ..configuration import Configuration
from ..constants import qpos_width
from .exceptions import LimitDefinitionError
from .limit import Constraint, Limit


class ConfigurationLimit(Limit):
    """Inequality constraint on joint positions in a robot model.

    Floating base joints are ignored.
    """

    def __init__(
        self,
        model: mujoco.MjModel,
        gain: float = 0.95,
        min_distance_from_limits: float = 0.0,
    ):
        """Initialize configuration limits.

        Args:
            model: MuJoCo model.
            gain: Gain factor in (0, 1] that determines how fast each joint is
                allowed to move towards the joint limits at each timestep. Values lower
                ttan 1 are safer but may make the joints move slowly.
            min_distance_from_limits: Offset in meters (slide joints) or radians
                (hinge joints) to be added to the limits. Positive values decrease the
                range of motion, negative values increase it (i.e. negative values
                allow penetration).
        """
        if not 0.0 < gain <= 1.0:
            raise LimitDefinitionError(
                f"{self.__class__.__name__} gain must be in the range (0, 1]"
            )

        index_list: list[int] = []
        lower = np.full(model.nq, -np.inf)
        upper = np.full(model.nq, np.inf)
        for jnt in range(model.njnt):
            jnt_type = model.jnt_type[jnt]
            qpos_dim = qpos_width(jnt_type)
            jnt_range = model.jnt_range[jnt]
            padr = model.jnt_qposadr[jnt]
            if jnt_type == mujoco.mjtJoint.mjJNT_FREE or not model.jnt_limited[jnt]:
                continue
            lower[padr : padr + qpos_dim] = jnt_range[0] + min_distance_from_limits
            upper[padr : padr + qpos_dim] = jnt_range[1] - min_distance_from_limits
            index_list.append(jnt)

        self.indices = np.array(index_list)
        self.indices.setflags(write=False)

        dim = len(self.indices)
        self.projection_matrix = np.eye(model.nv)[self.indices] if dim > 0 else None

        self.lower = lower
        self.upper = upper
        self.model = model
        self.gain = gain

    def compute_qp_inequalities(
        self,
        configuration: Configuration,
        dt: float,
    ) -> Constraint:
        r"""Compute the configuration-dependent joint position limits.

        The limits are defined as:

        .. math::

            {q \ominus q_{min}} \leq \Delta q \leq {q_{max} \ominus q}

        where :math:`q \in {\cal C}` is the robot's configuration and
        :math:`\Delta q \in T_q({\cal C})` is the displacement in the tangent
        space at :math:`q`. See the :ref:`derivations` section for more information.

        Args:
            configuration: Robot configuration :math:`q`.
            dt: Integration timestep in [s].

        Returns:
            Pair :math:`(G, h)` representing the inequality constraint as
            :math:`G \Delta q \leq h`, or ``None`` if there is no limit.
        """
        del dt  # Unused.

        # Upper.
        delta_q_max = np.zeros(self.model.nv)
        mujoco.mj_differentiatePos(
            m=self.model,
            qvel=delta_q_max,
            dt=1.0,
            qpos1=configuration.q,
            qpos2=self.upper,
        )

        # Lower.
        delta_q_min = np.zeros(self.model.nv)
        mujoco.mj_differentiatePos(
            m=self.model,
            qvel=delta_q_min,
            dt=1.0,
            # NOTE: mujoco.mj_differentiatePos does `qpos2 - qpos1` so notice the order
            # swap here compared to above.
            qpos1=self.lower,
            qpos2=configuration.q,
        )

        p_min = self.gain * delta_q_min[self.indices]
        p_max = self.gain * delta_q_max[self.indices]
        G = np.vstack([self.projection_matrix, -self.projection_matrix])
        h = np.hstack([p_max, p_min])
        return Constraint(G=G, h=h)



"""Kinematic tasks."""

from .com_task import ComTask
from .damping_task import DampingTask
from .exceptions import (
    InvalidDamping,
    InvalidGain,
    InvalidTarget,
    TargetNotSet,
    TaskDefinitionError,
)
from .frame_task import FrameTask
from .posture_task import PostureTask
from .relative_frame_task import RelativeFrameTask
from .relative_com_task import RelativeComTask
from .task import Objective, Task

__all__ = (
    "ComTask",
    "FrameTask",
    "Objective",
    "DampingTask",
    "PostureTask",
    "RelativeFrameTask",
    "Task",
    "TargetNotSet",
    "InvalidTarget",
    "TaskDefinitionError",
    "RelativeComTask",
    "InvalidGain",
    "InvalidDamping",
)



"""mink: MuJoCo inverse kinematics."""

from .configuration import Configuration
from .constants import (
    FRAME_TO_ENUM,
    FRAME_TO_JAC_FUNC,
    FRAME_TO_POS_ATTR,
    FRAME_TO_XMAT_ATTR,
    SUPPORTED_FRAMES,
)
from .exceptions import (
    InvalidFrame,
    InvalidKeyframe,
    InvalidMocapBody,
    MinkError,
    NotWithinConfigurationLimits,
    UnsupportedFrame,
)
from .lie import SE3, SO3, MatrixLieGroup
from .limits import (
    CollisionAvoidanceLimit,
    ConfigurationLimit,
    Constraint,
    Limit,
    VelocityLimit,
)
from .solve_ik import build_ik, solve_ik
from .tasks import (
    ComTask,
    DampingTask,
    FrameTask,
    Objective,
    PostureTask,
    RelativeFrameTask,
    RelativeComTask,
    TargetNotSet,
    Task,
)
from .utils import (
    custom_configuration_vector,
    get_body_geom_ids,
    get_freejoint_dims,
    get_subtree_geom_ids,
    move_mocap_to_frame,
)

__all__ = (
    "ComTask",
    "Configuration",
    "build_ik",
    "solve_ik",
    "DampingTask",
    "FrameTask",
    "RelativeFrameTask",
    "RelativeComTask",
    "PostureTask",
    "Task",
    "Objective",
    "ConfigurationLimit",
    "VelocityLimit",
    "CollisionAvoidanceLimit",
    "Constraint",
    "Limit",
    "SO3",
    "SE3",
    "MatrixLieGroup",
    "MinkError",
    "UnsupportedFrame",
    "InvalidFrame",
    "InvalidKeyframe",
    "NotWithinConfigurationLimits",
    "TargetNotSet",
    "InvalidMocapBody",
    "SUPPORTED_FRAMES",
    "FRAME_TO_ENUM",
    "FRAME_TO_JAC_FUNC",
    "FRAME_TO_POS_ATTR",
    "FRAME_TO_XMAT_ATTR",
    "set_mocap_pose_from_frame",
    "pose_from_mocap",
    "custom_configuration_vector",
    "get_freejoint_dims",
    "move_mocap_to_frame",
    "get_subtree_geom_ids",
    "get_body_geom_ids",
)




from dart_physics.robot_cfgs import * 
from pathlib import Path
import dart_physics.mink as mink 

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos_target): 

    ctrl = np.zeros(22)
    # ctrl
    # ctrl[:3] = qpos_target[0:3]
    # ctrl[3:9] = qpos_target[3:9] 
    # ctrl[3:10] = qpos_target[3:10]
    # ctrl[10:17] = qpos_target[12:19]
    # ctrl[0:3] = qpos_target[0:3]
    ctrl[0:6] = 0.0
    ctrl[6:13] = qpos_target[6:13] 
    ctrl[13:20] = qpos_target[15:22]
    
    # right gripper 
    # get from mocap 

    tip1 = mink.SE3.from_mocap_name(model, data, "right_rt_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "right_lt_target").as_matrix()
    dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[14] - 0.01
    ctrl[-2] = np.clip(dist - 0.01, max_grip, 0.05)
 

    # left gripper z
    # get from mocap 
    tip1 = mink.SE3.from_mocap_name(model, data, "left_rt_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "left_lt_target").as_matrix()
    dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[23] - 0.01
    ctrl[-1] = np.clip(dist - 0.01, max_grip, 0.05)
    # ctrl[20:24] = qpos_target[20:24]

    return ctrl 

joints = ["base_x", "base_y", "base_theta", "torso_0", 
          "torso_1", "torso_2", "torso_3", "torso_4", "torso_5", 
          "right_arm_0", "right_arm_1", "right_arm_2", "right_arm_3", "right_arm_4", "right_arm_5", "right_arm_6",
          "left_arm_0", "left_arm_1", "left_arm_2", "left_arm_3", "left_arm_4", "left_arm_5", "left_arm_6",
          "head_0", "head_1",
          "gripper_finger_r2", "gripper_finger_r1", "gripper_finger_l2", "gripper_finger_l1"]

bodies = {
    # "link_torso_0": "LINK_1", 
    # "link_torso_1": "LINK_2", 
    # "link_torso_2": "LINK_3", 
    # "link_torso_3": "LINK_4", 
    # "link_torso_4": "LINK_5", 
    # "link_torso_5": "LINK_6", 
    "link_right_arm_0": "LINK_7", 
    "link_right_arm_1": "LINK_8",
    "link_right_arm_2": "LINK_9",
    "link_right_arm_3": "LINK_10",
    "link_right_arm_4": "LINK_11",
    "link_right_arm_5": "LINK_12",
    "link_right_arm_6": "LINK_13",
    "FT_SENSOR_R": "FT_SENSOR_R", 
    "EE_BODY_R": "EE_BODY_RIGHT",  
    "ee_finger_r1": "EE_FINGER_RIGHT1",
    "ee_finger_r2": "EE_FINGER_RIGHT2",
    "link_left_arm_0": "LINK_14",
    "link_left_arm_1": "LINK_15",
    "link_left_arm_2": "LINK_16",
    "link_left_arm_3": "LINK_17",
    "link_left_arm_4": "LINK_18",
    "link_left_arm_5": "LINK_19",
    "link_left_arm_6": "LINK_20",
    "FT_SENSOR_L": "FT_SENSOR_L", 
    "EE_BODY_L": "EE_BODY_LEFT",  
    "ee_finger_l1": "EE_FINGER_LEFT1",
    "ee_finger_l2": "EE_FINGER_LEFT2",
}


robot_cfg = {

    "name": "RBY1_ONLY",

    "robots": { 
        "robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "rainbow_arm_only" / "model_act.xml").as_posix(),
                    home_q = [0] * 6 + [0, 0, 0, -1.46, 0, 0, 0] + [0, 0] + [0, 0, 0, -1.46, 0, 0, 0] + [0, 0] + [0, 0], #  +  
                    freejoint = False, 
                    attach_to = [0, -0.05, -0.8, np.sqrt(2)/2, 0, 0, np.sqrt(2)/2], 
                    parallel_jaw = True, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "ee_finger_l1", \
                                        body_name = "leftarm_right_finger_tip", pos = [0.0, 0.0, -0.05]),
                        NewBodyConfig(attach_to = "ee_finger_l2", \
                                        body_name = "leftarm_left_finger_tip", pos = [0, 0, -0.05]),

                        NewBodyConfig(attach_to="ee_finger_l1", \
                                        body_name = "leftarm_right_finger_base", pos = [0.03, 0, -0.02]),
                        NewBodyConfig(attach_to="ee_finger_l2", \
                                        body_name = "leftarm_left_finger_base", pos = [0.03, 0, -0.02]),

                        NewBodyConfig(attach_to = "ee_finger_r1", \
                                        body_name = "rightarm_right_finger_tip", pos = [0.0, 0.0, -0.05]),
                        NewBodyConfig(attach_to = "ee_finger_r2", \
                                        body_name = "rightarm_left_finger_tip", pos = [0.0, 0.0, -0.05]),

                        NewBodyConfig(attach_to = "ee_finger_r1", \
                                        body_name = "rightarm_right_finger_base", pos = [0.03, 0, -0.02]),
                        NewBodyConfig(attach_to = "ee_finger_r2", \
                                        body_name = "rightarm_left_finger_base", pos = [0.03, 0, -0.02]),

                        NewBodyConfig(attach_to = "link_head_1", \
                                        body_name = "target_head", pos = [0.0, 0.0, 0.0]),
                    ], 
                    
                ), 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": 26, 
        
    "points" : { 

        "left_lb":      FingerKeypoints("left", 3, "left_lb_target", "robot/leftarm_left_finger_base", type = "body"), 
        "left_lt":      FingerKeypoints("left", 4, "left_lt_target", "robot/leftarm_left_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("left", 8, "left_rb_target", "robot/leftarm_right_finger_base", type = "body"),
        "left_rt":      FingerKeypoints("left", 9, "left_rt_target", "robot/leftarm_right_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("right", 3, "right_lb_target", "robot/rightarm_left_finger_base", type = "body"),
        "right_lt":     FingerKeypoints("right", 4, "right_lt_target", "robot/rightarm_left_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("right", 8, "right_rb_target", "robot/rightarm_right_finger_base", type = "body"),
        "right_rt":     FingerKeypoints("right", 9, "right_rt_target", "robot/rightarm_right_finger_tip", type = "body"),

        # "target_head":  FingerKeypoints("head", 1, "target_head", "robot/target_head", type = "body"),

    },

    "reverse_points": { 

        "left_lb":      FingerKeypoints("right", 3, "left_lb_target", "robot/leftarm_right_finger_base", type = "body"),
        "left_lt":      FingerKeypoints("right", 4, "left_lt_target", "robot/leftarm_right_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("right", 8, "left_rb_target", "robot/leftarm_left_finger_base", type = "body"),
        "left_rt":      FingerKeypoints("right", 9, "left_rt_target", "robot/leftarm_left_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("left", 3, "right_lb_target", "robot/rightarm_right_finger_base", type = "body"),
        "right_lt":     FingerKeypoints("left", 4, "right_lt_target", "robot/rightarm_right_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("left", 8, "right_rb_target", "robot/rightarm_left_finger_base", type = "body"),
        "right_rt":     FingerKeypoints("left", 9, "right_rt_target", "robot/rightarm_left_finger_tip", type = "body"),

        # "target_head":  FingerKeypoints("head", 1, "target_head", "robot/target_head", type = "body"),
    },

    "avp_calib": { 

        "left_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_lt": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rt": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

        "right_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_lt": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rt": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

        # "target_head": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(target = "left_lb"),
        IKTasks(target = "left_lt"),
        IKTasks(target = "left_rb"),
        IKTasks(target = "left_rt"),

        IKTasks(target = "right_lb"),
        IKTasks(target = "right_lt"),
        IKTasks(target = "right_rb"),
        IKTasks(target = "right_rt"),
    ],

    "regulation_task": [ 
        # PostureTask(qpos=[0] * 26, cost = 0.01, disable_joints = [9, 10, 11, 12, 13, 14, 15, 16] + [18, 19, 20, 21, 22, 23, 24, 25],), 
        # RelativeCoMTask(root_name = "robot/base", root_type = "body", cost = 1.0),
        # DampingTask(cost = 1.0, disable_joints = [3, 4, 5, 6, 7, 8] + [10, 11, 12, 13, 14, 15, 16] + [18, 19, 20, 21, 22, 23, 24, 25])
    ], 

    "joints" : [f'robot/{j}' for j in joints],

    "bodies":  {f"robot/{b}": v for b, v in bodies.items()}, 
    
}   


if __name__ == "__main__": 

    # import trimesh 
    # from scipy.spatial.transform import Rotation as R
    # import os 

    # asset_root = os.path.join(_HERE, "../assets", "aloha", "assets")
    # usdz_root = os.path.join(_HERE, "../assets", "aloha", "transformed_stl")\

    # os.makedirs(usdz_root, exist_ok = True)

    # print([f"{k}" for k in robot_cfg["bodies"].values()])

    # for k, v in robot_cfg["usdz_geoms"].items():

    #     meshes = [] 
    #     for i, geom in enumerate(v): 
    #         geom: GeomInfo
    #         mesh = trimesh.load_mesh(f"{asset_root}/{geom.mesh}.stl")
    #         pos = np.array(geom.pos)
    #         quat = np.array(geom.quat)

    #         mesh.visual.vertex_colors = np.array([0, 0, 0, 255], dtype = np.uint8)

    #         mat = np.eye(4)
    #         mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
    #         mat[:3, 3] = pos

    #         mesh.apply_transform(mat)

    #         meshes.append(mesh)

    #     scale = [mesh.scale for mesh in meshes]
    #     avg_scale = np.mean(scale, axis = 0)
    #     print(k, avg_scale)
        
    #     meshes = trimesh.util.concatenate(meshes)

    #     if avg_scale > 100: 

    #         meshes.apply_scale(0.001)

    #     if avg_scale < 1: 

    #         meshes.apply_scale(1)
        
    #     meshes.export(f"{usdz_root}/{k}.glb")


    # print the "values" of bodies, 
    #  i need the values to be in "x", "y", "z" format 
    # i explicitly need the quoatation marks

    strings = [] 
    for k, v in robot_cfg["bodies"].items(): 
        strings.append(f'"{v}"')  # Corrected the formatting here

    print(",".join(strings))



import mujoco
import mujoco.viewer
from dm_control import mjcf
import numpy as np
from loop_rate_limiters import RateLimiter

import math 
import numpy as np 
import time 
import dart_physics.mink as mink
import subprocess 
import os 
import yaml 
from scipy.spatial.transform import Rotation as R
from flask import Flask, request, jsonify
import threading 
import sys 
import matplotlib.pyplot as plt 
import zmq 
from queue import Queue
from threading import Thread
from dirs import * 

# import below modules from .. path 
import sys
sys.path.append("./")
from utils.scene_gen import * 
from utils.mesh_download import download_mesh
from utils.scene_gen import USDZ_LINK_CORRESPONDENCE
from utils.opt_ik import PandaOptIK
from copy import deepcopy 
import torch 
import multiprocessing as mp

def load_cfg(path): 
    with open(path, "r") as f: 
        return yaml.load(f, Loader = yaml.FullLoader)

cfg = load_cfg(f"cfgs/mug_hang.yaml")



class BatchRenderer: 
    def __init__(self, cfg): 
        
        self.model = construct_dual_panda(cfg) 
        self.data = mujoco.MjData(self.model) 

        self.renderer = mujoco.Renderer(self.model, 480, 640)

    def render(self, qpos): 

        """
        qpos:  torch.Tensor(n_batch, qpos.shape)
        return: torch.Tensor(n_batch, 480, 640, 3)
        """

        qpos = qpos.cpu().numpy()
        imgs = [] 
        for i in range(qpos.shape[0]): 
            self.data.qpos[:] = qpos[i]
            self.renderer.update_scene(self.data, camera = 'head_cam')
            image = self.renderer.render()
            
            imgs.append(image)
        imgs = np.stack(imgs, axis = 0) 
        imgs = torch.from_numpy(imgs)

        return imgs



class MPBatchRenderer:
    def __init__(self, cfg):
        self.cfg = cfg
        self.model = construct_dual_panda(self.cfg)
        self.data = mujoco.MjData(self.model)

    def _initialize_renderer(self):
        """
        Initializes the model, data, and renderer for each process.
        This method is called once per process when it starts.
        """
        self.model = construct_dual_panda(self.cfg)
        self.data = mujoco.MjData(self.model)
        self.renderer = mujoco.Renderer(self.model, 480, 640)

    def _render_chunk(self, qpos_chunk):
        """
        Render a chunk of qpos instances.
        """
        imgs = []
        for qpos in qpos_chunk:
            self.data.qpos[:] = qpos
            self.renderer.update_scene(self.data, camera='head_cam')
            image = self.renderer.render()
            imgs.append(image)
        return imgs

    def render(self, qpos, num_workers=8):
        """
        qpos: torch.Tensor(n_batch, qpos.shape)
        return: torch.Tensor(n_batch, 480, 640, 3)
        """

        qpos = qpos.cpu().numpy()
        qpos_chunks = np.array_split(qpos, num_workers)

        with mp.Pool(processes=num_workers, initializer=self._initialize_renderer) as pool:
            results = pool.map(self._render_chunk, qpos_chunks)

        # Flatten the list of images
        imgs = [img for sublist in results for img in sublist]
        imgs = np.stack(imgs, axis=0)
        imgs = torch.from_numpy(imgs)

        return imgs


if __name__ == "__main__": 

    br = MPBatchRenderer(cfg) 

    qpos_size = br.data.qpos.shape[0]
    
    bs = 128 
    times = [] 
    for i in range(10): 
        t0 = time.time() 
        qpos =  torch.randn((bs, qpos_size))
        imgs = br.render(qpos)
        t1 = time.time()
        times.append(t1 - t0)
        print("Time taken to render {} images: {}".format(bs, t1 - t0))
    print("Average time: ", np.mean(times))
        
        






from dart_physics.robot_cfgs import * 
from pathlib import Path
import dart_physics.mink as mink 

_HERE = Path(__file__).resolve().parent


def qpos2ctrl(model, data, qpos_target): 

    ctrl = np.zeros(14)
    ctrl[:6] = qpos_target[0:6]
    ctrl[7:13] = qpos_target[8:14]

    # left gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "left_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "left_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[7] - 0.01
    ctrl[6] = np.clip(finger_dist - 0.01, max_grip, 0.04)

    # right gripper control
    tip1 = mink.SE3.from_mocap_name(model, data, "right_lf_target").as_matrix()
    tip2 = mink.SE3.from_mocap_name(model, data, "right_rf_target").as_matrix()

    finger_dist = np.linalg.norm(tip1[:3, 3] - tip2[:3, 3])
    max_grip = data.qpos[15] - 0.01
    ctrl[13] = np.clip(finger_dist - 0.01, max_grip, 0.04)


    return ctrl 

joints = ["left_waist", "left_shoulder", "left_elbow", "left_forearm_roll", 
          "left_wrist_angle", "left_wrist_rotate", "left_left_finger", "left_right_finger",
          "right_waist", "right_shoulder", "right_elbow", "right_forearm_roll",
        "right_wrist_angle", "right_wrist_rotate", "right_left_finger", "right_right_finger"]

bodies = ["left_base_link", "left_shoulder_link", "left_upper_arm_link", "left_upper_forearm_link",
          "left_lower_forearm_link", "left_wrist_link", "left_gripper_base", "left_left_finger_link", 
          "left_right_finger_link"] + \
        ["right_base_link", "right_shoulder_link", "right_upper_arm_link", "right_upper_forearm_link",
        "right_lower_forearm_link", "right_wrist_link", "right_gripper_base", "right_left_finger_link", 
        "right_right_finger_link"]


robot_cfg = {

    "name": "ALOHA",

    "robots": { 
        "robot": RobotConfig( 
                    xml_path = (_HERE / "../assets" / "aloha" / "aloha.xml").as_posix(),
                    home_q = [0, -0.96, 1.16, 0, -0.3, 0, 0.0084, 0.0084, 0, -0.96, 1.16, 0, -0.3, 0, 0.0084, 0.0084], 
                    freejoint = False, 
                    attach_to = [0.0, 0.2, 0.0, 1, 0, 0, 0 ], 
                    parallel_jaw = True, 
                    add_bodies = [ 
                        NewBodyConfig(attach_to = "left_right_finger_link", \
                                      body_name = "leftarm_right_finger_base", pos = [0.0, 0.0, 0.0]),
                        NewBodyConfig(attach_to = "left_left_finger_link", \
                                        body_name = "leftarm_left_finger_base", pos = [0, 0, 0.0]), 
                        NewBodyConfig(attach_to = "left_right_finger_link", \
                                        body_name = "leftarm_right_finger_tip", pos = [0.0, 0.05, 0.02]),
                        NewBodyConfig(attach_to = "left_left_finger_link", \
                                        body_name = "leftarm_left_finger_tip", pos = [0, -0.05, 0.02]),

                        NewBodyConfig(attach_to = "right_right_finger_link", \
                                        body_name = "rightarm_right_finger_base", pos = [0.0, 0.0, 0.0]),
                        NewBodyConfig(attach_to = "right_left_finger_link", \
                                        body_name = "rightarm_left_finger_base", pos = [0, 0, 0.0]),
                        NewBodyConfig(attach_to = "right_right_finger_link", \
                                        body_name = "rightarm_right_finger_tip", pos = [0.0, 0.05, 0.02]),
                        NewBodyConfig(attach_to = "right_left_finger_link", \
                                        body_name = "rightarm_left_finger_tip", pos = [0, -0.05, 0.02]),
                                      
                    ], 
                    
                ), 
    }, 

    "qpos2ctrl": qpos2ctrl, 
    "obj_startidx": (7+1) * 2, 
        
    "points" : { 

        "left_lb":     FingerKeypoints("left", 3, "left_lb_target", "robot/leftarm_left_finger_base", type = "body"),
        "left_lf":      FingerKeypoints("left", 4, "left_lf_target", "robot/leftarm_left_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("left", 8, "left_rb_target", "robot/leftarm_right_finger_base", type = "body"),
        "left_rf":      FingerKeypoints("left", 9, "left_rf_target", "robot/leftarm_right_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("right", 8, "right_lb_target", "robot/rightarm_left_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("right", 9, "right_lf_target", "robot/rightarm_left_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("right", 3, "right_rb_target", "robot/rightarm_right_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("right", 4, "right_rf_target", "robot/rightarm_right_finger_tip", type = "body"),

    },

    "reverse_points": { 

        "left_lb":     FingerKeypoints("right", 3, "left_lb_target", "robot/leftarm_right_finger_base", type = "body"),
        "left_lf":      FingerKeypoints("right", 4, "left_lf_target", "robot/leftarm_right_finger_tip", type = "body"),
        "left_rb":      FingerKeypoints("right", 8, "left_rb_target", "robot/leftarm_left_finger_base", type = "body"),
        "left_rf":      FingerKeypoints("right", 9, "left_rf_target", "robot/leftarm_left_finger_tip", type = "body"),
        
        "right_lb":     FingerKeypoints("left", 8, "right_lb_target", "robot/rightarm_right_finger_base", type = "body"),
        "right_lf":     FingerKeypoints("left", 9, "right_lf_target", "robot/rightarm_right_finger_tip", type = "body"),
        "right_rb":     FingerKeypoints("left", 3, "right_rb_target", "robot/rightarm_left_finger_base", type = "body"),
        "right_rf":     FingerKeypoints("left", 4, "right_rf_target", "robot/rightarm_left_finger_tip", type = "body"),
    },

    "avp_calib": { 

        "left_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "left_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

        "right_lb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_lf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rb": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},
        "right_rf": {"scale": 1.0, "offset": np.array([0.0, 0.0, 0.0])},

    }, 

    "ik_task" : [
        IKTasks(target = "left_lb"), 
        IKTasks(target = "left_lf"),
        IKTasks(target = "left_rb"),
        IKTasks(target = "left_rf"),

        IKTasks(target = "right_lb"),
        IKTasks(target = "right_lf"),
        IKTasks(target = "right_rb"),
        IKTasks(target = "right_rf"),
    ],

    "joints" : [f'robot/{j}' for j in joints],

    "bodies":  {f"robot/{b}": f"final_aloha_{b}" for b in bodies}, 

    "cameras": [
        CameraInfo("stereo_cam", pos = [-0.09433973, -0.29207777,  0.69795567], \
                   euler = [41.44131086, 0.5, 1.84107336], fovy = 70.0 ), 
    ], 

    "usdz_geoms": { 
        "aloha_left_base_link": [
            GeomInfo(mesh = "vx300s_1_base", quat = [1, 0, 0, 1]),
        ], 
        "aloha_left_shoulder_link": [
            GeomInfo(mesh = "vx300s_2_shoulder", pos = [0, 0, -0.003], quat = [1, 0, 0, 1]),
        ],
        "aloha_left_upper_arm_link": [ 
            GeomInfo(mesh = "vx300s_3_upper_arm", quat = [1, 0, 0, 1]),
        ],
        "aloha_left_upper_forearm_link": [ 
            GeomInfo(mesh = "vx300s_4_upper_forearm"), 
        ],
        "aloha_left_lower_forearm_link": [ 
            GeomInfo(mesh = "vx300s_5_lower_forearm", quat = [0, 1, 0, 0]), 
        ],
        "aloha_left_wrist_link": [ 
            GeomInfo(mesh = "vx300s_6_wrist", quat = [1, 0, 0, 1]), 
        ],
        "aloha_left_gripper_base": [ 
            GeomInfo(mesh = "vx300s_7_gripper_prop"), 
            GeomInfo(mesh = "vx300s_7_gripper_bar"),
            GeomInfo(mesh = "vx300s_7_gripper_wrist_mount", pos=[0, -0.03525, -0.0227], quat=[0, -1, 0, -1]),
            GeomInfo(mesh = "d405_solid", pos=[0, -0.0824748, -0.0095955], quat=[0, 0, -0.21644, -0.976296]), 
        ],
        "aloha_left_left_finger_link": [ 
            GeomInfo(mesh = "vx300s_8_custom_finger_left", pos=[0.0141637, 0.0211727, 0.06], quat=[1, 1, 1, -1]),    
        ],
        "aloha_left_right_finger_link": [
            GeomInfo(mesh = "vx300s_8_custom_finger_right", pos=[0.0141637, -0.0211727, 0.0597067], quat=[1, -1, -1, -1]),
        ],

        "aloha_right_base_link": [
            GeomInfo(mesh = "vx300s_1_base", quat = [1, 0, 0, 1]),
        ],
        "aloha_right_shoulder_link": [
            GeomInfo(mesh = "vx300s_2_shoulder", pos = [0, 0, -0.003], quat = [1, 0, 0, 1]),
        ],
        "aloha_right_upper_arm_link": [ 
            GeomInfo(mesh = "vx300s_3_upper_arm", quat = [1, 0, 0, 1]),
        ],
        "aloha_right_upper_forearm_link": [ 
            GeomInfo(mesh = "vx300s_4_upper_forearm"), 
        ],
        "aloha_right_lower_forearm_link": [ 
            GeomInfo(mesh = "vx300s_5_lower_forearm", quat = [0, 1, 0, 0]), 
        ],
        "aloha_right_wrist_link": [ 
            GeomInfo(mesh = "vx300s_6_wrist", quat = [1, 0, 0, 1]), 
        ],
        "aloha_right_gripper_base": [ 
            GeomInfo(mesh = "vx300s_7_gripper_prop"), 
            GeomInfo(mesh = "vx300s_7_gripper_bar"),
            GeomInfo(mesh = "vx300s_7_gripper_wrist_mount", pos=[0, -0.03525, -0.0227], quat=[0, -0.707107, 0, -0.707107]),
            GeomInfo(mesh = "d405_solid", pos=[0, -0.0824748, -0.0095955], quat=[0, 0, -0.21644, -0.976296]), 
        ],
        "aloha_right_left_finger_link": [ 
            GeomInfo(mesh = "vx300s_8_custom_finger_left", pos=[0.0141637, 0.0211727, 0.06], quat=[1, 1, 1, -1]),    
        ],
        "aloha_right_right_finger_link": [
            GeomInfo(mesh = "vx300s_8_custom_finger_right", pos=[0.0141637, -0.0211727, 0.0597067], quat=[1, -1, -1, -1]),
        ],
    }

}   


if __name__ == "__main__": 

    import trimesh 
    from scipy.spatial.transform import Rotation as R
    import os 

    asset_root = os.path.join(_HERE, "../assets", "aloha", "assets")
    usdz_root = os.path.join(_HERE, "../assets", "aloha", "transformed_stl")\

    os.makedirs(usdz_root, exist_ok = True)

    print([f"{k}" for k in robot_cfg["bodies"].values()])

    for k, v in robot_cfg["usdz_geoms"].items():

        meshes = [] 
        for i, geom in enumerate(v): 
            geom: GeomInfo
            mesh = trimesh.load_mesh(f"{asset_root}/{geom.mesh}.stl")
            pos = np.array(geom.pos)
            quat = np.array(geom.quat)

            mesh.visual.vertex_colors = np.array([0, 0, 0, 255], dtype = np.uint8)

            mat = np.eye(4)
            mat[:3, :3] = R.from_quat(quat, scalar_first = True).as_matrix()
            mat[:3, 3] = pos

            mesh.apply_transform(mat)

            meshes.append(mesh)

        scale = [mesh.scale for mesh in meshes]
        avg_scale = np.mean(scale, axis = 0)
        print(k, avg_scale)
        
        meshes = trimesh.util.concatenate(meshes)

        if avg_scale > 100: 

            meshes.apply_scale(0.001)

        if avg_scale < 1: 

            meshes.apply_scale(1)
        
        meshes.export(f"{usdz_root}/{k}.glb")





import isaacgym
import torch 
from isaacgym import gymapi
import numpy as np
import argparse
import torch.nn.functional as F
from typing import * 
import yaml 
import os 

asset_loaded = False 

def load_cfg(cfg_path: str) -> Dict: 
    cfg = yaml.load(open(cfg_path, 'r'), Loader=yaml.FullLoader)
    return cfg 

def default_sim_params(use_gpu, up_axis = 'Z', hz = 60.0, substeps = 4, num_position_iterations = 8, num_velocity_iterations = 2):
    sim_params = gymapi.SimParams()
    sim_params.up_axis = gymapi.UP_AXIS_Y if up_axis == 'Y' else gymapi.UP_AXIS_Z
    sim_params.gravity = gymapi.Vec3(0.0, -9.8, 0.0) if up_axis == 'Y' else gymapi.Vec3(0.0, 0.0, -9.8)
    sim_params.dt = 1.0 / hz
    sim_params.substeps = substeps
    sim_params.use_gpu_pipeline = use_gpu 
    sim_params.physx.solver_type = 1
    sim_params.physx.num_position_iterations = num_position_iterations
    sim_params.physx.num_velocity_iterations = num_velocity_iterations
    sim_params.physx.rest_offset = 0.0
    sim_params.physx.contact_offset = 0.001
    sim_params.physx.friction_offset_threshold = 0.001
    sim_params.physx.friction_correlation_distance = 0.0005
    sim_params.physx.num_threads = 4
    sim_params.physx.use_gpu = use_gpu
    return sim_params


def load_axis(gym, sim, device, size, asset_root = './assets'):

    robot_asset_file = "{}_axis.urdf".format(size)
    asset_options = gymapi.AssetOptions()
    asset_options.armature = 0.01
    asset_options.fix_base_link = True
    asset_options.disable_gravity = True
    robot_asset = gym.load_asset(sim, asset_root, robot_asset_file, asset_options)

    return robot_asset



def load_left_leap_hand_asset(gym, sim, device, asset_root = '../bidex_sim/assets'):

    robot_asset_file = "robots/hands/allegro_hand/allegro_hand_left.urdf"
    asset_options = gymapi.AssetOptions()
    asset_options.armature = 0.01
    asset_options.fix_base_link = False
    asset_options.disable_gravity = True
    asset_options.flip_visual_attachments = False
    asset_options.use_mesh_materials = True
    asset_options.vhacd_enabled = True
    asset_options.vhacd_params = gymapi.VhacdParams()
    asset_options.vhacd_params.resolution = 1000000
    robot_asset = gym.load_asset(sim, asset_root, robot_asset_file, asset_options)

    return robot_asset 

def load_ur3e_asset(gym, sim, device, asset_root = '../bidex_sim/assets', hand = None, chirality = None, control = {'arm': 'POS',  'hand': 'POS'}):

    if hand is None: 
        robot_asset_file = "robots/ur_description/urdf/ur3e.urdf"
    else:
        robot_asset_file = "robots/ur_description/urdf/ur3e_{}_{}.urdf".format(hand, chirality)

    asset_options = gymapi.AssetOptions()
    asset_options.armature = 0.01
    asset_options.fix_base_link = True
    asset_options.disable_gravity = True
    asset_options.flip_visual_attachments = False
    asset_options.use_mesh_materials = True
    asset_options.vhacd_enabled = True
    asset_options.vhacd_params = gymapi.VhacdParams()
    if control['arm'] == 'POS': 
        asset_options.default_dof_drive_mode = gymapi.DOF_MODE_POS
    elif control['arm'] == 'EFFORT':
        asset_options.default_dof_drive_mode = gymapi.DOF_MODE_EFFORT
    asset_options.vhacd_params.resolution = 1000000
    robot_asset = gym.load_asset(sim, asset_root, robot_asset_file, asset_options)
    robot_dof_props = gym.get_asset_dof_properties(robot_asset)
    for i in range(6, 22):
        robot_dof_props['driveMode'][i] = gymapi.DOF_MODE_POS
        robot_dof_props['stiffness'][i] = 10000
        robot_dof_props['damping'][i] = 500
    return robot_asset, robot_dof_props


def refresh_tensors(gym, sim): 
    # refresh tensors
    gym.refresh_rigid_body_state_tensor(sim)
    gym.refresh_actor_root_state_tensor(sim)
    gym.refresh_dof_state_tensor(sim)
    gym.refresh_jacobian_tensors(sim)
    gym.refresh_mass_matrix_tensors(sim)

def setup_viewer_camera(gym, env, viewer): 
    cam_pos = gymapi.Vec3(0.0, -0.05, 1.55)
    cam_target = gymapi.Vec3(0.0, 0.5, 0.4)
    middle_env = env
    gym.viewer_camera_look_at(viewer, middle_env, cam_pos, cam_target)

def adjust_viewer_camera(gym, env, viewer, cam_pos, cam_target):
    gym.viewer_camera_look_at(viewer, env, cam_pos, cam_target)


import trimesh
import xml.etree.ElementTree as ET
from scipy.spatial.transform import Rotation as R
import os
import mujoco
from dm_control import mjcf

# Load the MuJoCo XML model
xml_path = "./assets/wonik_allegro"
xml_file = 'left_hand.xml'

xml_full_path = os.path.join(xml_path, xml_file)

# Load the MJCF model
model = mjcf.from_path(xml_full_path)

def apply_transformations(mesh, pos, quat):
    """Apply translation and rotation to a mesh."""
    # Apply translation
    translation_matrix = trimesh.transformations.translation_matrix(pos)
    # Apply rotation
    rotation_matrix = R.from_quat(quat).as_matrix()
    rotation_matrix_4x4 = trimesh.transformations.identity_matrix()
    rotation_matrix_4x4[:3, :3] = rotation_matrix
    # Apply transformation
    transformation_matrix = trimesh.transformations.concatenate_matrices(translation_matrix, rotation_matrix_4x4)
    return mesh.apply_transform(transformation_matrix)

def process_body(body):
    """Process a body, apply transformations, and export its mesh."""
    body_name = body.full_identifier.replace("/", "_")
    meshes = []
    
    for geom in body.find_all('geom'):
        pos = geom.pos if geom.pos is not None else [0.0, 0.0, 0.0]
        quat = geom.quat if geom.quat is not None else [1.0, 0.0, 0.0, 0.0]
        mesh_file = None
        
        if geom.mesh:
            print(geom.mesh.name)
            mesh_file = geom.mesh.name
        # else:
        #     print(geom.dclass)
        elif geom.dclass and geom.dclass.geom.mesh:
            mesh_file = geom.dclass.geom.mesh.name
        
        if mesh_file:
            mesh_file_path = os.path.join(xml_path, 'assets', mesh_file + '.stl')
            if os.path.exists(mesh_file_path):
                mesh = trimesh.load(mesh_file_path)
                # Apply the transformations
                transformed_mesh = apply_transformations(mesh, pos, quat)
                meshes.append(transformed_mesh)
            else:
                print(f"Warning: Mesh file {mesh_file_path} not found.")
    
    # Recursively process child bodies
    for child_body in body.all_children():
        child_meshes = process_body(child_body)
        meshes.extend(child_meshes)
    
    # Combine all meshes under this body, even if it is empty
    # if not meshes:
    #     # Create an empty mesh if no geometries are found
    #     empty_mesh = trimesh.Trimesh(vertices=[], faces=[])
    #     meshes.append(empty_mesh)
    if meshes == []: 
        return meshes
    combined_mesh = trimesh.util.concatenate(meshes)
    output_path = os.path.join(xml_path, f'{body_name}.stl')
    combined_mesh.export(output_path)

    return meshes

# Start processing from the root body
root_body = model.worldbody
process_body(root_body)




import torch
import torch.nn.functional as F 
import numpy as np
from typing import * 

def standardize_quaternion(quaternions: torch.Tensor) -> torch.Tensor:
    """
    Convert a unit quaternion to a standard form: one in which the real
    part is non negative.

    Args:
        quaternions: Quaternions with real part first,
            as tensor of shape (..., 4).

    Returns:
        Standardized quaternions as tensor of shape (..., 4).
    """
    return torch.where(quaternions[..., 0:1] < 0, -quaternions, quaternions)


def _sqrt_positive_part(x: torch.Tensor) -> torch.Tensor:
    """
    Returns torch.sqrt(torch.max(0, x))
    but with a zero subgradient where x is 0.
    """
    ret = torch.zeros_like(x)
    positive_mask = x > 0
    ret[positive_mask] = torch.sqrt(x[positive_mask])
    return ret


def matrix_to_quaternion(matrix: torch.Tensor) -> torch.Tensor:
    """
    Convert rotations given as rotation matrices to quaternions.

    Args:
        matrix: Rotation matrices as tensor of shape (..., 3, 3).

    Returns:
        quaternions with real part first, as tensor of shape (..., 4).
    """
    if matrix.size(-1) != 3 or matrix.size(-2) != 3:
        raise ValueError(f"Invalid rotation matrix shape {matrix.shape}.")

    batch_dim = matrix.shape[:-2]
    m00, m01, m02, m10, m11, m12, m20, m21, m22 = torch.unbind(
        matrix.reshape(batch_dim + (9,)), dim=-1
    )

    q_abs = _sqrt_positive_part(
        torch.stack(
            [
                1.0 + m00 + m11 + m22,
                1.0 + m00 - m11 - m22,
                1.0 - m00 + m11 - m22,
                1.0 - m00 - m11 + m22,
            ],
            dim=-1,
        )
    )

    # we produce the desired quaternion multiplied by each of r, i, j, k
    quat_by_rijk = torch.stack(
        [
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack([q_abs[..., 0] ** 2, m21 - m12, m02 - m20, m10 - m01], dim=-1),
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack([m21 - m12, q_abs[..., 1] ** 2, m10 + m01, m02 + m20], dim=-1),
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack([m02 - m20, m10 + m01, q_abs[..., 2] ** 2, m12 + m21], dim=-1),
            # pyre-fixme[58]: `**` is not supported for operand types `Tensor` and
            #  `int`.
            torch.stack([m10 - m01, m20 + m02, m21 + m12, q_abs[..., 3] ** 2], dim=-1),
        ],
        dim=-2,
    )

    # We floor here at 0.1 but the exact level is not important; if q_abs is small,
    # the candidate won't be picked.
    flr = torch.tensor(0.1).to(dtype=q_abs.dtype, device=q_abs.device)
    quat_candidates = quat_by_rijk / (2.0 * q_abs[..., None].max(flr))

    # if not for numerical problems, quat_candidates[i] should be same (up to a sign),
    # forall i; we pick the best-conditioned one (with the largest denominator)
    out = quat_candidates[
        F.one_hot(q_abs.argmax(dim=-1), num_classes=4) > 0.5, :
    ].reshape(batch_dim + (4,))
    return standardize_quaternion(out)

def quaternion_to_matrix(quaternions: torch.Tensor) -> torch.Tensor:
    """
    Convert rotations given as quaternions to rotation matrices.

    Args:
        quaternions: quaternions with real part first,
            as tensor of shape (..., 4).

    Returns:
        Rotation matrices as tensor of shape (..., 3, 3).
    """
    r, i, j, k = torch.unbind(quaternions, -1)
    # pyre-fixme[58]: `/` is not supported for operand types `float` and `Tensor`.
    two_s = 2.0 / (quaternions * quaternions).sum(-1)

    o = torch.stack(
        (
            1 - two_s * (j * j + k * k),
            two_s * (i * j - k * r),
            two_s * (i * k + j * r),
            two_s * (i * j + k * r),
            1 - two_s * (i * i + k * k),
            two_s * (j * k - i * r),
            two_s * (i * k - j * r),
            two_s * (j * k + i * r),
            1 - two_s * (i * i + j * j),
        ),
        -1,
    )
    return o.reshape(quaternions.shape[:-1] + (3, 3))



def isaac_mat2quat(mat: torch.Tensor) -> torch.Tensor:
    """
    returns posquat in pos + xyzw format (following isaacgym convention)
    """
    wxyz = matrix_to_quaternion(mat)
    xyzw = torch.cat([wxyz[:, 1:], wxyz[:, :1]], dim=-1)
    return xyzw

def mat2posquat(mat: torch.Tensor) -> torch.Tensor:
    pos = mat[..., :3, 3]
    quat = isaac_mat2quat(mat[..., :3, :3])
    return torch.cat([pos, quat], dim=-1)

def posquat2mat(posquat: torch.Tensor) -> torch.Tensor:
    """
    convert pos + quaternion in xyzw format to matrix 
    """
    batch = posquat.shape[0]
    pos = posquat[..., :3]
    quat_xyzw = posquat[..., 3:]
    quat_wxyz = torch.cat([quat_xyzw[:, 3:], quat_xyzw[:, :3]], dim=-1)
    rot_mat = quaternion_to_matrix(quat_wxyz)

    result = torch.eye(4, device=pos.device, dtype=pos.dtype).unsqueeze(0).expand(batch, -1, -1)
    result[..., :3, :3] = rot_mat
    result[..., :3, 3] = pos

    return result




"""Relative center-of-mass task implementation."""

from __future__ import annotations

from typing import Optional

import mujoco
import numpy as np
import numpy.typing as npt

from ..configuration import Configuration
from .exceptions import InvalidTarget, TargetNotSet, TaskDefinitionError
from .task import Task


class RelativeComTask(Task):
    """Regulate the center-of-mass (CoM) of a robot relative to another frame.

    Attributes:
        root_name: Name of the root frame.
        root_type: The root frame type: 'body', 'geom', or 'site'.
        target_com_in_root_frame: Target CoM position expressed in the root frame.
    """

    k: int = 3
    target_com_in_root_frame: Optional[np.ndarray]

    def __init__(
        self,
        root_name: str,
        root_type: str,
        cost: npt.ArrayLike,
        gain: float = 1.0,
        lm_damping: float = 0.0,
    ):
        super().__init__(cost=np.zeros((self.k,)), gain=gain, lm_damping=lm_damping)
        self.root_name = root_name
        self.root_type = root_type
        self.target_com_in_root_frame = None

        self.set_cost(cost)

    def set_cost(self, cost: npt.ArrayLike) -> None:
        cost = np.atleast_1d(cost)
        if cost.ndim != 1 or cost.shape[0] not in (1, self.k):
            raise TaskDefinitionError(
                f"{self.__class__.__name__} cost must be a vector of shape (1,) "
                f"(aka identical cost for all coordinates) or ({self.k},). "
                f"Got {cost.shape}"
            )
        if not np.all(cost >= 0.0):
            raise TaskDefinitionError(f"{self.__class__.__name__} cost must be >= 0")
        self.cost[:] = cost

    def set_target(self, target_com_in_root_frame: npt.ArrayLike) -> None:
        """Set the target CoM position in the root frame.

        Args:
            target_com_in_root_frame: Desired center-of-mass position in the root frame.
        """
        target_com_in_root_frame = np.atleast_1d(target_com_in_root_frame)
        if target_com_in_root_frame.ndim != 1 or target_com_in_root_frame.shape[0] != (self.k):
            raise InvalidTarget(
                f"Expected target CoM in root frame to have shape ({self.k},) but got "
                f"{target_com_in_root_frame.shape}"
            )
        self.target_com_in_root_frame = target_com_in_root_frame.copy()

    def set_target_from_configuration(self, configuration: Configuration) -> None:
        """Set the target CoM position in the root frame from a given robot configuration.

        Args:
            configuration: Robot configuration :math:`q`.
        """
        # Get CoM position in world frame
        current_com = configuration.data.subtree_com[1]
        # Get transform from root frame to world frame
        transform_root_to_world = configuration.get_transform_frame_to_world(
            self.root_name, self.root_type
        )
        # Get inverse transform (world to root frame)
        transform_world_to_root = transform_root_to_world.inverse()
        # Compute current CoM in root frame coordinates
        current_com_in_root_frame = transform_world_to_root.apply(current_com)
        # Set target_com_in_root_frame
        self.set_target(current_com_in_root_frame)

    def compute_error(self, configuration: Configuration) -> np.ndarray:
        """Compute the CoM task error relative to the root frame.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Center-of-mass task error vector :math:`e(q)`.
        """
        if self.target_com_in_root_frame is None:
            raise TargetNotSet(self.__class__.__name__)

        # Get CoM position in world frame
        current_com = configuration.data.subtree_com[1]
        # Get transform from root frame to world frame
        transform_root_to_world = configuration.get_transform_frame_to_world(
            self.root_name, self.root_type
        )
        # Get inverse transform (world to root frame)
        transform_world_to_root = transform_root_to_world.inverse()
        # Compute current CoM in root frame coordinates
        current_com_in_root_frame = transform_world_to_root.apply(current_com)
        # Compute error
        error = current_com_in_root_frame - self.target_com_in_root_frame
        return error

    def compute_jacobian(self, configuration: Configuration) -> np.ndarray:
        """Compute the CoM task Jacobian relative to the root frame.

        Args:
            configuration: Robot configuration :math:`q`.

        Returns:
            Center-of-mass task Jacobian :math:`J(q)`.
        """
        if self.target_com_in_root_frame is None:
            raise TargetNotSet(self.__class__.__name__)

        # Get CoM Jacobian in world frame
        J_com = np.empty((self.k, configuration.nv))
        mujoco.mj_jacSubtreeCom(configuration.model, configuration.data, J_com, 1)

        # Get Jacobian of root frame position in world frame
        J_root_full = configuration.get_frame_jacobian(self.root_name, self.root_type)
        J_root_pos = J_root_full[:3, :]  # First 3 rows: position Jacobian

        # Get rotation matrix from world frame to root frame
        transform_root_to_world = configuration.get_transform_frame_to_world(
            self.root_name, self.root_type
        )
        R_wr = transform_root_to_world.rotation().as_matrix()
        R_rw = R_wr.T  # Transpose to get world to root frame rotation

        # Compute Jacobian of CoM in root frame coordinates
        J_com_root = R_rw @ (J_com - J_root_pos)

        return J_com_root



from typing import Optional

import mujoco
import numpy as np

from . import constants as consts
from .exceptions import InvalidKeyframe, InvalidMocapBody


def move_mocap_to_frame(
    model: mujoco.MjModel,
    data: mujoco.MjData,
    mocap_name: str,
    frame_name: str,
    frame_type: str,
) -> None:
    """Initialize mocap body pose at a desired frame.

    Args:
        model: Mujoco model.
        data: Mujoco data.
        mocap_name: The name of the mocap body.
        frame_name: The desired frame name.
        frame_type: The desired frame type. Can be "body", "geom" or "site".
    """
    mocap_id = model.body(mocap_name).mocapid[0]
    if mocap_id == -1:
        raise InvalidMocapBody(mocap_name, model)

    obj_id = mujoco.mj_name2id(model, consts.FRAME_TO_ENUM[frame_type], frame_name)
    xpos = getattr(data, consts.FRAME_TO_POS_ATTR[frame_type])[obj_id]
    xmat = getattr(data, consts.FRAME_TO_XMAT_ATTR[frame_type])[obj_id]

    data.mocap_pos[mocap_id] = xpos.copy()
    mujoco.mju_mat2Quat(data.mocap_quat[mocap_id], xmat)


def get_freejoint_dims(model: mujoco.MjModel) -> tuple[list[int], list[int]]:
    """Get all floating joint configuration and tangent indices.

    Args:
        model: Mujoco model.

    Returns:
        A (q_ids, v_ids) pair containing all floating joint indices in the
        configuration and tangent spaces respectively.
    """
    q_ids: list[int] = []
    v_ids: list[int] = []
    for j in range(model.njnt):
        if model.jnt_type[j] == mujoco.mjtJoint.mjJNT_FREE:
            qadr = model.jnt_qposadr[j]
            vadr = model.jnt_dofadr[j]
            q_ids.extend(range(qadr, qadr + 7))
            v_ids.extend(range(vadr, vadr + 6))
    return q_ids, v_ids


def custom_configuration_vector(
    model: mujoco.MjModel,
    key_name: Optional[str] = None,
    **kwargs,
) -> np.ndarray:
    """Generate a configuration vector where named joints have specific values.

    Args:
        model: Mujoco model.
        key_name: Optional keyframe name to initialize the configuration vector from.
            Otherwise, the default pose `qpos0` is used.
        kwargs: Custom values for joint coordinates.

    Returns:
        Configuration vector where named joints have the values specified in
            keyword arguments, and other joints have their neutral value or value
            defined in the keyframe if provided.
    """
    data = mujoco.MjData(model)
    if key_name is not None:
        key_id = mujoco.mj_name2id(model, mujoco.mjtObj.mjOBJ_KEY, key_name)
        if key_id == -1:
            raise InvalidKeyframe(key_name, model)
        mujoco.mj_resetDataKeyframe(model, data, key_id)
    else:
        mujoco.mj_resetData(model, data)
    q = data.qpos.copy()
    for name, value in kwargs.items():
        jid = model.joint(name).id
        jnt_dim = consts.qpos_width(model.jnt_type[jid])
        qid = model.jnt_qposadr[jid]
        value = np.atleast_1d(value)
        if value.shape != (jnt_dim,):
            raise ValueError(
                f"Joint {name} should have a qpos value of {jnt_dim,} but "
                f"got {value.shape}"
            )
        q[qid : qid + jnt_dim] = value
    return q


def get_subtree_geom_ids(model: mujoco.MjModel, body_id: int) -> list[int]:
    """Get all geoms belonging to subtree starting at a given body.

    Args:
        model: Mujoco model.
        body_id: ID of body where subtree starts.

    Returns:
        A list containing all subtree geom ids.
    """

    def gather_geoms(body_id: int) -> list[int]:
        geoms: list[int] = []
        geom_start = model.body_geomadr[body_id]
        geom_end = geom_start + model.body_geomnum[body_id]
        geoms.extend(range(geom_start, geom_end))
        children = [i for i in range(model.nbody) if model.body_parentid[i] == body_id]
        for child_id in children:
            geoms.extend(gather_geoms(child_id))
        return geoms

    return gather_geoms(body_id)


def get_body_geom_ids(model: mujoco.MjModel, body_id: int) -> list[int]:
    """Get all geoms belonging to a given body.

    Args:
        model: Mujoco model.
        body_id: ID of body.

    Returns:
        A list containing all body geom ids.
    """
    geom_start = model.body_geomadr[body_id]
    geom_end = geom_start + model.body_geomnum[body_id]
    return list(range(geom_start, geom_end))



from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple

import mujoco
import numpy as np

from .base import MatrixLieGroup
from .utils import get_epsilon, skew

_IDENTITIY_WXYZ = np.array([1.0, 0.0, 0.0, 0.0], dtype=np.float64)
_INVERT_QUAT_SIGN = np.array([1.0, -1.0, -1.0, -1.0], dtype=np.float64)


class RollPitchYaw(NamedTuple):
    """Struct containing roll, pitch, and yaw Euler angles."""

    roll: float
    pitch: float
    yaw: float


@dataclass(frozen=True)
class SO3(MatrixLieGroup):
    """Special orthogonal group for 3D rotations.

    Internal parameterization is (qw, qx, qy, qz). Tangent parameterization is
    (omega_x, omega_y, omega_z).
    """

    wxyz: np.ndarray
    matrix_dim: int = 3
    parameters_dim: int = 4
    tangent_dim: int = 3
    space_dim: int = 3

    def __post_init__(self) -> None:
        if self.wxyz.shape != (self.parameters_dim,):
            raise ValueError(
                f"Expeced wxyz to be a length 4 vector but got {self.wxyz.shape[0]}."
            )

    def __repr__(self) -> str:
        wxyz = np.round(self.wxyz, 5)
        return f"{self.__class__.__name__}(wxyz={wxyz})"

    def parameters(self) -> np.ndarray:
        return self.wxyz

    def copy(self) -> SO3:
        return SO3(wxyz=self.wxyz.copy())

    @classmethod
    def from_x_radians(cls, theta: float) -> SO3:
        return SO3.exp(np.array([theta, 0.0, 0.0], dtype=np.float64))

    @classmethod
    def from_y_radians(cls, theta: float) -> SO3:
        return SO3.exp(np.array([0.0, theta, 0.0], dtype=np.float64))

    @classmethod
    def from_z_radians(cls, theta: float) -> SO3:
        return SO3.exp(np.array([0.0, 0.0, theta], dtype=np.float64))

    @classmethod
    def from_rpy_radians(
        cls,
        roll: float,
        pitch: float,
        yaw: float,
    ) -> SO3:
        return (
            SO3.from_z_radians(yaw)
            @ SO3.from_y_radians(pitch)
            @ SO3.from_x_radians(roll)
        )

    @classmethod
    def from_matrix(cls, matrix: np.ndarray) -> SO3:
        assert matrix.shape == (SO3.matrix_dim, SO3.matrix_dim)
        wxyz = np.zeros(SO3.parameters_dim, dtype=np.float64)
        mujoco.mju_mat2Quat(wxyz, matrix.ravel())
        return SO3(wxyz=wxyz)

    @classmethod
    def identity(cls) -> SO3:
        return SO3(wxyz=_IDENTITIY_WXYZ)

    @classmethod
    def sample_uniform(cls) -> SO3:
        # Ref: https://lavalle.pl/planning/node198.html
        u1, u2, u3 = np.random.uniform(
            low=np.zeros(shape=(3,)),
            high=np.array([1.0, 2.0 * np.pi, 2.0 * np.pi]),
        )
        a = np.sqrt(1.0 - u1)
        b = np.sqrt(u1)
        wxyz = np.array(
            [
                a * np.sin(u2),
                a * np.cos(u2),
                b * np.sin(u3),
                b * np.cos(u3),
            ],
            dtype=np.float64,
        )
        return SO3(wxyz=wxyz)

    # Eq. 138.
    def as_matrix(self) -> np.ndarray:
        mat = np.zeros(9, dtype=np.float64)
        mujoco.mju_quat2Mat(mat, self.wxyz)
        return mat.reshape(3, 3)

    def compute_roll_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arctan2(2 * (q0 * q1 + q2 * q3), 1 - 2 * (q1**2 + q2**2))

    def compute_pitch_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arcsin(2 * (q0 * q2 - q3 * q1))

    def compute_yaw_radians(self) -> float:
        q0, q1, q2, q3 = self.wxyz
        return np.arctan2(2 * (q0 * q3 + q1 * q2), 1 - 2 * (q2**2 + q3**2))

    def as_rpy_radians(self) -> RollPitchYaw:
        return RollPitchYaw(
            roll=self.compute_roll_radians(),
            pitch=self.compute_pitch_radians(),
            yaw=self.compute_yaw_radians(),
        )

    # Paragraph above Appendix B.A.
    def inverse(self) -> SO3:
        return SO3(wxyz=self.wxyz * _INVERT_QUAT_SIGN)

    def normalize(self) -> SO3:
        return SO3(wxyz=self.wxyz / np.linalg.norm(self.wxyz))

    # Eq. 136.
    def apply(self, target: np.ndarray) -> np.ndarray:
        assert target.shape == (SO3.space_dim,)
        padded_target = np.concatenate([np.zeros(1, dtype=np.float64), target])
        return (self @ SO3(wxyz=padded_target) @ self.inverse()).wxyz[1:]

    def multiply(self, other: SO3) -> SO3:
        res = np.empty(self.parameters_dim, dtype=np.float64)
        mujoco.mju_mulQuat(res, self.wxyz, other.wxyz)
        return SO3(wxyz=res)

    ##
    #
    ##

    # Eq. 132.
    @classmethod
    def exp(cls, tangent: np.ndarray) -> SO3:
        assert tangent.shape == (SO3.tangent_dim,)
        theta_squared = tangent @ tangent
        theta_pow_4 = theta_squared * theta_squared
        use_taylor = theta_squared < get_epsilon(tangent.dtype)
        safe_theta = 1.0 if use_taylor else np.sqrt(theta_squared)
        safe_half_theta = 0.5 * safe_theta
        if use_taylor:
            real = 1.0 - theta_squared / 8.0 + theta_pow_4 / 384.0
            imaginary = 0.5 - theta_squared / 48.0 + theta_pow_4 / 3840.0
        else:
            real = np.cos(safe_half_theta)
            imaginary = np.sin(safe_half_theta) / safe_theta
        wxyz = np.concatenate([np.array([real]), imaginary * tangent])
        return SO3(wxyz=wxyz)

    # Eq. 133.
    def log(self) -> np.ndarray:
        w = self.wxyz[0]
        norm_sq = self.wxyz[1:] @ self.wxyz[1:]
        use_taylor = norm_sq < get_epsilon(norm_sq.dtype)
        norm_safe = 1.0 if use_taylor else np.sqrt(norm_sq)
        w_safe = w if use_taylor else 1.0
        atan_n_over_w = np.arctan2(-norm_safe if w < 0 else norm_safe, abs(w))
        if use_taylor:
            atan_factor = 2.0 / w_safe - 2.0 / 3.0 * norm_sq / w_safe**3
        else:
            if abs(w) < get_epsilon(w.dtype):
                scl = 1.0 if w > 0.0 else -1.0
                atan_factor = scl * np.pi / norm_safe
            else:
                atan_factor = 2.0 * atan_n_over_w / norm_safe
        return atan_factor * self.wxyz[1:]

    # Eq. 139.
    def adjoint(self) -> np.ndarray:
        return self.as_matrix()

    # Jacobians.

    # Eqn. 145, 174.
    @classmethod
    def ljac(cls, other: np.ndarray) -> np.ndarray:
        theta = np.sqrt(other @ other)
        use_taylor = theta < get_epsilon(theta.dtype)
        if use_taylor:
            t2 = theta**2
            A = (1.0 / 2.0) * (1.0 - t2 / 12.0 * (1.0 - t2 / 30.0 * (1.0 - t2 / 56.0)))
            B = (1.0 / 6.0) * (1.0 - t2 / 20.0 * (1.0 - t2 / 42.0 * (1.0 - t2 / 72.0)))
        else:
            A = (1 - np.cos(theta)) / (theta**2)
            B = (theta - np.sin(theta)) / (theta**3)
        skew_other = skew(other)
        return np.eye(3) + A * skew_other + B * (skew_other @ skew_other)

    @classmethod
    def ljacinv(cls, other: np.ndarray) -> np.ndarray:
        theta = np.sqrt(other @ other)
        use_taylor = theta < get_epsilon(theta.dtype)
        if use_taylor:
            t2 = theta**2
            A = (1.0 / 12.0) * (1.0 + t2 / 60.0 * (1.0 + t2 / 42.0 * (1.0 + t2 / 40.0)))
        else:
            A = (1.0 / theta**2) * (
                1.0 - (theta * np.sin(theta) / (2.0 * (1.0 - np.cos(theta))))
            )
        skew_other = skew(other)
        return np.eye(3) - 0.5 * skew_other + A * (skew_other @ skew_other)



import os 
from pathlib import Path

CUR_PATH = Path(__file__).resolve().parent

LOG_STORE_DIR = os.getenv('LOG_STORE_DIR', os.path.join(CUR_PATH, 'logs'))
ASSET_STORE_DIR = os.getenv('ASSETS_STORE_DIR', os.path.join(CUR_PATH, 'assets', 'custom')) #f'{CUR_PATH}/assets/custom')
MJB_PATH = os.getenv('ASSETS_STORE_DIR', os.path.join(CUR_PATH, 'assets', 'custom')) 


import dart_physics.avp_stream3.grpc_msg.handtracking_pb2 as handtracking_pb2
import dart_physics.avp_stream3.grpc_msg.handtracking_pb2_grpc as handtracking_pb2_grpc
import dart_physics.avp_stream3.grpc_msg.dexterity_pb2 as dexterity_pb2
import dart_physics.avp_stream3.grpc_msg.dexterity_pb2_grpc as dexterity_pb2_grpc


from pathlib import Path

import mujoco
import mujoco._structs
import mujoco.viewer
from dm_control import mjcf
import numpy as np
from loop_rate_limiters import RateLimiter

import math 
import numpy as np 
import time 
import dart_physics.mink as mink
import subprocess 
import os 
import yaml 
from scipy.spatial.transform import Rotation as R
from flask import Flask, request, jsonify
import threading 
import sys 
import matplotlib.pyplot as plt 
import zmq 
from queue import Queue
from threading import Thread
from dirs import * 

# import below modules from .. path 
import sys
sys.path.append("./")
from utils.scene_gen import construct_scene
from copy import deepcopy 
from PIL import Image 
from typing import * 

from dart_physics.utils.run_functions import *
from utils.diff_ik import DiffIK
from avp_stream3 import HandTrackingServer 
from runs import load_task_cfg, load_robot_cfg
from utils.aloha_zmq_sub import ALOHAZMQSubscriber
from utils.camera_viewer import * 
import warnings
warnings.filterwarnings("ignore")


control_message = None 
app = Flask(__name__)

def run_flask():
    app.run(host='0.0.0.0', port=7013)


def keyboard_listener(): 
    global control_message  
    print(f"R - Start recording\nS - Save recording\nC - Cancel recording\nD - Delete last recording\nO - Reset objects\nQ - Quit")
    while True:
        # Check if there is input available
        key = sys.stdin.read(1).strip().lower()

        if key == 's':
            # start recording 
            control_message = "start_recording"
        elif key == 'd': 
            # stop recording (without saving)
            control_message = "stop_recording"

        elif key == 'f': 
            # save recording
            control_message = "save" 

        elif key == 'q':
            control_message = "reset_scene"

        elif key == 'w': 
            control_message = "reset_robot"
            
        elif key == 't':
            control_message = "switch_sm"


@app.route('/control', methods=['POST'])
def control():
    global control_message

    data = request.json
    control_message = data['command']

    return jsonify(success=True, message=f"Received command: {control_message}")

dummy_cfg = load_task_cfg("dummy")

class MJTeleop: 
    
    def __init__(self, task, robot, headless = False, avp = False): 

        self.cfg = load_task_cfg(task)
        self.robot_cfg = load_robot_cfg(robot)
        self.headless = headless
        self.avp = avp

        self.profiler = time_profiler()

        model = construct_scene(self.cfg, self.robot_cfg)
        self.configuration = mink.Configuration(model)
        self.data = self.configuration.data
        self.model = self.configuration.model


        model = construct_scene(dummy_cfg, self.robot_cfg)
        self.dummy_configuration = mink.Configuration(model)
        self.dummy_data = self.dummy_configuration.data
        self.dummy_model = self.dummy_configuration.model

        if not self.headless:
            self.viewer = mujoco.viewer.launch_passive(
                model=self.model, data=self.data, show_left_ui=True, show_right_ui=True)


        self.ik = DiffIK(self.dummy_configuration, self.robot_cfg)

        self.mocap_indices = dict() 
        for key, value in self.robot_cfg['points'].items():
            self.mocap_indices[value.mocap_frame] = self.model.body(value.mocap_frame).mocapid[0]

        # "head_track" mocap 
        self.mocap_indices["stereo_cam"] = self.model.body("stereo_cam").mocapid[0]

        self.rate = RateLimiter(frequency=500.0)
        self.t = 0.0

        self.is_recording = False 
        self.recording_start_time = 0.0
        self.trajectory = [] 
        self.hri_cmd = None 

        self.meta_data = { 
            "task": self.cfg["name"],
            "robot": self.robot_cfg["name"],
            "objects": self.cfg['objects'],
        }

        self.init = False

        self.prev_dict = None 

        if args.render: 
            if args.stereo: 
                self.renderer = mujoco.Renderer(self.model, 240, 640)
            else:
                self.renderer = mujoco.Renderer(self.model, 240, 320)

            self.zmq_addr = "tcp://127.0.0.1:5555"
            camera_viewer_process = start_camera_viewer_process(self.zmq_addr, args.stereo)
            if args.stereo: 
                self.renderer.scene.stereo = mujoco.mjtStereo.mjSTEREO_SIDEBYSIDE

            self.socket = get_socket(self.zmq_addr)
            

        # Start Flask server in a separate thread
        threading.Thread(target=run_flask, daemon=True).start()
        threading.Thread(target=keyboard_listener, daemon=True).start()

        self.streamer = HandTrackingServer()
        self.streamer.start()

        if args.puppet: 
            self.zmq_subscriber = ALOHAZMQSubscriber(zmq_port = 5555)
            self.zmq_subscriber.start()
        

    def preprocess_targets(self): 

        for key, value in self.robot_cfg["points"].items():
            
            if value.weld: 

                pos = value.init_posquat[:3]
                quat = value.init_posquat[3:]

                self.data.mocap_pos[self.model.body(value.mocap_frame).mocapid[0]] = np.array(pos)
                self.data.mocap_quat[self.model.body(value.mocap_frame).mocapid[0]] = np.array(quat)

            else:
                mink.move_mocap_to_frame(self.model, self.data, \
                                         value.mocap_frame, value.body_frame, value.type)
    
            

    def preprocess_avp(self, latest): 

        avp_frames = dict.fromkeys(["left", "right"], []) 

        d = deepcopy(latest)

        for key, transforms in self.robot_cfg['avp_calib'].items():

            finger = self.robot_cfg['points'][key]
            avp_idx = finger.avp_idx
            chilarity = finger.chilarity

            scale = transforms['scale']
            offset = transforms['offset']

            d[chilarity + "_fingers"][avp_idx][:3, -1] *= scale
            d[chilarity + "_fingers"][avp_idx][:3, -1] += offset

        avp_frames["right"] = d['right_wrist'] @ d['right_fingers']
        avp_frames["right"][:, 2, -1] -= args.th 

        avp_frames["left"]  = d['left_wrist'] @ d['left_fingers']
        avp_frames["left"][:, 2, -1] -= args.th 


        required_frames = dict.fromkeys(["left", "right"], {})

        for key, finger in self.robot_cfg['points'].items():

            chilarity = finger.chilarity
            avp_idx = finger.avp_idx
            mocap_frame = finger.mocap_frame
            rot_transform = finger.avp_transform

            f = avp_frames[chilarity][avp_idx]
            f[:3, :3] = f[:3, :3] @ rot_transform[:3, :3]

            required_frames[chilarity] [mocap_frame] = f 

        required_frames["head"] = d['head']
        required_frames["head"][:, 2, -1] -= ( args.th )

        return required_frames


    def check_if_exploding(self, threshold = 5):

        # print max data.qvel 
        print("Max qvel", np.max(np.abs(self.data.qvel)))
        if np.any(np.abs(self.data.qvel) > threshold):
            
            return True

        return False
    

    def solve_ik(self): 

        for task_desc, task in zip (self.robot_cfg["ik_task"], self.ik.tasks) :

            if task_desc.type == "relative":

                root = task_desc.root
                target = task_desc.target

                frame1 = self.robot_cfg["points"][root].body_frame
                frame2 = self.robot_cfg["points"][target].mocap_frame

                T_pm = self.configuration.get_transform(
                    frame2, "body", frame1, "body"
                )
                task.set_target(T_pm)                    

            elif task_desc.type == "absolute":

                target = task_desc.target

                frame = self.robot_cfg["points"][target].mocap_frame
                task.set_target(mink.SE3.from_mocap_name(self.model, self.data, frame))
        
        qtarget = self.ik.solve()

        return qtarget 



    def run(self): 

        mujoco.mj_resetDataKeyframe(self.model, self.data, self.model.key("home").id)

        mujoco.mj_step(self.model, self.data)

        self.preprocess_targets()

        if not self.init: 
            self.init_data = deepcopy(self.data)
            self.init = True

        cnt = 0 

        while self.viewer.is_running() if not args.headless else True:

            self.profiler.start("total")

            self.profiler.start("avp_preprocess")

            if self.avp: 

                latest = self.streamer.get_latest()
                if latest is None: 
                    continue

                processed_avp_frames = self.preprocess_avp(latest)

                for key, value in processed_avp_frames["right"].items():
                    self.data.mocap_pos[self.mocap_indices[key]] = value[:3, -1]
                    self.data.mocap_quat[self.mocap_indices[key]] = R.from_matrix(value[:3, :3]).as_quat(scalar_first = True)

                for key, value in processed_avp_frames["left"].items():
                    self.data.mocap_pos[self.mocap_indices[key]] = value[:3, -1]
                    self.data.mocap_quat[self.mocap_indices[key]] = R.from_matrix(value[:3, :3]).as_quat(scalar_first = True)

                if args.active: 
                    # stereo_cam 
                    self.data.mocap_pos[self.mocap_indices["stereo_cam"]] = processed_avp_frames["head"][0, :3, -1]
                    self.data.mocap_quat[self.mocap_indices["stereo_cam"]] = R.from_matrix(processed_avp_frames["head"][0, :3, :3]).as_quat(scalar_first = True)

                self.hri_cmd = latest 

            self.profiler.end("avp_preprocess")

            self.profiler.start("ik solving")
            if cnt % (500 // args.freq) == 0: 
                
                if not args.puppet: 
                    qtarget = self.solve_ik()

            if args.puppet: 
                if self.zmq_subscriber.latest is not None:
                    latest = self.zmq_subscriber.latest
                    self.data.ctrl[:6] = latest[:6]
                    self.data.ctrl[6] = latest[7]
                    self.data.ctrl[7:13] = latest[9:15]
                    self.data.ctrl[13] = latest[16]

            else:
                self.data.ctrl = self.robot_cfg['qpos2ctrl'](self.model, self.data, qtarget)

            
            self.profiler.end("ik solving")

            self.profiler.start("physics step")
            mujoco.mj_step(self.model, self.data)

            self.ik.update(self.data.qpos[:self.robot_cfg["obj_startidx"]])
            self.profiler.end("physics step")

            if not self.headless:
                self.viewer.sync()

            self.profiler.start("post physics control")
            self.post_physics_control()
            self.profiler.end("post physics control")


            if control_message is None: 

                self.profiler.start("dictionary preparing")
                sim_dict = self.prepare_dict_to_send_avp() 
                self.streamer.set_sim_states(sim_dict) 
                self.prev_dict = deepcopy(sim_dict) 
                self.profiler.end("dictionary preparing")

            if args.render:

                if cnt % (500 // 20) == 0: 

                    self.renderer.update_scene(self.data, "stereo_cam")
                    img = self.renderer.render()
                    self.socket.send_pyobj(img) 

                    print("setting image from sim")
                    self.streamer.set_image_states(img, args.stereo)
                    print("setted image from sim")

            self.profiler.end("total")

            cnt += 1 
            self.rate.sleep()

    def randomize_wrist_cam(self):

        body = self.model.body("l_robot/wrist")
        body.pos = np.array([-0.04, 0, -0.05]) #+ np.random.uniform(-0.005, 0.005, (3,))
        euler = np.array([0, -1.2217305, 0]) 
        euler += np.random.uniform(-0.1, 0.1, 3)
        mat = R.from_euler("xyz", euler, degrees = False).as_matrix() 
        mat = mat @ rot_y(-90)[:3, :3] @ rot_z(-90)[:3, :3]
        quat = R.from_matrix(mat).as_quat(scalar_first = True)
        body.quat = quat
        # euler += np.random.uniform(-0.1, 0.1, 3)
        # body.quat = R.from_euler('xyz', euler, degrees = False).as_quat(scalar_first = True)


    def post_physics_control(self): 
        global control_message

        if self.is_recording:
            state_dict = self.prepare_dict_to_record()
            self.trajectory.append(state_dict)

        if control_message == "reset_robot": 

            self.data.qpos = self.init_data.qpos
            self.data.qvel = self.init_data.qvel
            self.data.ctrl = self.init_data.ctrl
            self.data.mocap_pos = self.init_data.mocap_pos
            self.data.mocap_quat = self.init_data.mocap_quat

            mujoco.mj_forward(self.model, self.data)

            if self.check_if_exploding(): 
                print("Exploding!")
            else:
                control_message = None 

            print(f"[RECORDING CTRL] Resetting Robot.") 

        if control_message == "reset_scene":
            

            reset_func = self.cfg.get('reset_function', None)

            if reset_func is not None:
                reset_func(self.model, self.data, self.robot_cfg, self.cfg)

            else:

                id = self.robot_cfg["obj_startidx"] # TODO : fix this
                vid = self.robot_cfg["obj_startidx"] 
            
                try: 
                    reset_cfg = self.cfg.get('reset_perturbation', {})
                except:
                    pass 

                        

                for ii, obj in enumerate(self.cfg['objects']): 

                    try:  
                        if not obj in reset_cfg:
                            print(f"[WARN] Object {obj} does not specify how to perturb its pose. Skipping...")
                            new_posquat = np.array(self.cfg['default_poses'][obj]).copy()
                        else:
                            new_posquat = get_random_perturbed_pose(
                                old_pose=np.array(self.cfg['default_poses'][obj]),
                                reset_spec=reset_cfg[obj],
                            )
                        
                        self.data.qpos[id:id+7] = new_posquat
                        self.data.qvel[vid:vid+6] = 0.0

                    except:
                        rand_range = self.cfg['randomize_range']
                        posquat = np.array(self.cfg['default_poses'][obj]) 
                        self.data.qpos[id:id+7] = posquat

                        if self.cfg['randomize_pos'][obj]:
                            self.data.qpos[id] += np.random.uniform(-rand_range, rand_range)
                            self.data.qpos[id+1] += np.random.uniform(-rand_range, rand_range)

                        if self.cfg['randomize_rot'][obj]:
                            rand_quat = np.random.uniform(-1, 1, 4)
                            rand_quat /= np.linalg.norm(rand_quat)
                            self.data.qpos[id+3:id+7] = rand_quat

                        self.data.qvel[vid:vid+6] = 0.0

                    id += 7 
                    vid += 6

            mujoco.mj_step(self.model, self.data)
            self.preprocess_targets()

            if self.check_if_exploding(): 
                print("Exploding!")
            else:
                control_message = None 

            print(f"[RECORDING CTRL] Resetting Scene") 


        if control_message == "start_recording": 
            self.is_recording = True 
            self.recording_start_time = time.time()
            control_message = None

            print(f"[RECORDING CTRL] Started recording.")


        if control_message == "save":
            self.is_recording = False 
            control_message = None

            # see how many files are there
            files = [f for f in os.listdir(f"{LOG_STORE_DIR}") if f.endswith(".npy")] 
            timestamp = time.strftime("%Y/%m/%d-%H:%M:%S")

            time_elapsed = time.time() - self.recording_start_time
            self.meta_data['time_elapsed'] = time_elapsed
            self.meta_data['time_stamp'] = timestamp
            self.meta_data['success'] = True 

            # save in :03d format 
            np.save(f"{LOG_STORE_DIR}/{len(files):03d}.npy", self.trajectory)
            # save meta data in yaml format
            with open(f"{LOG_STORE_DIR}/{len(files):03d}.yaml", "w") as f: 
                yaml.dump(self.meta_data, f)

            self.trajectory = [] 
            print(f"[RECORDING CTRL] Saved successful trajectory to logs/{len(files):03d}.npy")


        if control_message == "stop_recording":
            self.is_recording = False 
            control_message = None

            # see how many files are there
            files = [f for f in os.listdir(f"{LOG_STORE_DIR}") if f.endswith(".npy")] 
            timestamp = time.strftime("%Y/%m/%d-%H:%M:%S")

            time_elapsed = time.time() - self.recording_start_time
            self.meta_data['time_elapsed'] = time_elapsed
            self.meta_data['time_stamp'] = timestamp
            self.meta_data['success'] = False  

            # save in :03d format 
            np.save(f"{LOG_STORE_DIR}/{len(files):03d}.npy", self.trajectory)
            # save meta data in yaml format
            with open(f"{LOG_STORE_DIR}/{len(files):03d}.yaml", "w") as f: 
                yaml.dump(self.meta_data, f)

            self.trajectory = [] 
            print("[RECORDING CTRL] Recording saved as failure.")

        if self.data.time == 0: 

            mujoco.mj_resetDataKeyframe(self.model, self.data, self.model.key("home").id)
            mujoco.mj_forward(self.model, self.data)
            self.preprocess_targets()

        if self.data.warning[mujoco.mjtWarning.mjWARN_BADQACC].number > 0:
            
            print("warning: bad qacc")
            mujoco.mj_resetDataKeyframe(self.model, self.data, self.model.key("home").id)
            mujoco.mj_forward(self.model, self.data)
            self.preprocess_targets()


    def prepare_dict_to_send_avp(self): 

        # this is a dictionary that I send to vision pro for AR rendering. 

        state_dict = {}

        for body_name, usdz_name in self.robot_cfg["bodies"].items(): 
            _body = self.data.body(body_name)
            state_dict[usdz_name] = mink.SE3.from_rotation_and_translation(rotation = mink.SO3(_body.xquat), translation = _body.xpos).as_matrix()

        for obj in self.cfg['objects']: 
            body_lists = self.cfg.get('object_meshes', {}) 

            if obj not in body_lists.keys(): 
                _obj = self.data.body(obj + '/')
                state_dict[obj] = mink.SE3.from_rotation_and_translation(rotation = mink.SO3(_obj.xquat), translation = _obj.xpos).as_matrix() 
            else: 
                body_list = body_lists[obj] 
                for body in body_list: 
                    _obj = self.data.body(obj + '/' + body)
                    state_dict[obj + '_' + body] = mink.SE3.from_rotation_and_translation(rotation = mink.SO3(_obj.xquat), translation = _obj.xpos).as_matrix()
                
        for v in state_dict.values():
            v[2, -1] += args.th

        return state_dict 

    def compare_dicts(self, prev_dict, new_dict): 

        if prev_dict is None: 
            return new_dict
        else:
            # compare diffeernces for every items, and only send the keys that are "different" to the server
            diff_dict = {}
            for key in prev_dict.keys():
                diff = np.linalg.norm(prev_dict[key] - new_dict[key])
                if diff > 1e-4:
                    diff_dict[key] = new_dict[key]

            return diff_dict


    def prepare_dict_to_record(self): 

        # need to prepare a dictionary containing simulation states to send to server 
        
        state_dict = {}

        for obj in self.cfg["objects"]: 
            _obj = self.data.body(obj + '/')
            state_dict[obj] = mink.SE3.from_rotation_and_translation(rotation = mink.SO3(_obj.xquat), translation = _obj.xpos).as_matrix()

        state_dict['entire_qpos'] = deepcopy(self.data.qpos)
        state_dict['entire_qvel'] = deepcopy(self.data.qvel)

        state_dict['qpos'] = deepcopy(self.data.qpos[..., :18])
        state_dict['qvel'] = deepcopy(self.data.qvel[..., :18])
        state_dict['ctrl'] = deepcopy(self.data.ctrl)
        state_dict['hri_cmd'] = self.hri_cmd

        return state_dict 


if __name__ == "__main__":
    
    import argparse 

    parser = argparse.ArgumentParser()
    parser.add_argument("--robot", default = "mug_hang")
    parser.add_argument("--task", default = "mug_hang")
    parser.add_argument("--headless", action = 'store_true')
    parser.add_argument("--avp", action = 'store_true')
    parser.add_argument("--th", type = float, default = 0.75)
    parser.add_argument("--render", action = 'store_true')
    parser.add_argument("--puppet", action = 'store_true')
    parser.add_argument("--stereo", action = 'store_true')
    parser.add_argument("--active", action = 'store_true')
    # choose control frequency, choice of 10, 20, 25, 50 
    parser.add_argument("--freq", type = int, default = 100, choices = [10, 20, 25, 50, 100, 500])
    args = parser.parse_args()    

    teleop = MJTeleop(args.task, args.robot, args.headless, args.avp)
    teleop.run()


import abc
from typing import Union, overload

import numpy as np
from typing_extensions import Self


class MatrixLieGroup(abc.ABC):
    """Interface definition for matrix Lie groups.

    Attributes:
        matrix_dim: Dimension of square matrix output.
        parameters_dim: Dimension of underlying parameters.
        tangent_dim: Dimension of tangent space.
        space_dim: Dimension of coordinates that can be transformed.
    """

    matrix_dim: int
    parameters_dim: int
    tangent_dim: int
    space_dim: int

    @overload
    def __matmul__(self, other: Self) -> Self: ...

    @overload
    def __matmul__(self, other: np.ndarray) -> np.ndarray: ...

    def __matmul__(self, other: Union[Self, np.ndarray]) -> Union[Self, np.ndarray]:
        """Overload of the @ operator."""
        if isinstance(other, np.ndarray):
            return self.apply(target=other)
        assert isinstance(other, MatrixLieGroup)
        return self.multiply(other=other)

    # Factory.

    @classmethod
    @abc.abstractmethod
    def identity(cls) -> Self:
        """Returns identity element."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_matrix(cls, matrix: np.ndarray) -> Self:
        """Get group member from matrix representation."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def sample_uniform(cls) -> Self:
        """Draw a uniform sample from the group."""
        raise NotImplementedError

    # Accessors.

    @abc.abstractmethod
    def as_matrix(self) -> np.ndarray:
        """Get transformation as a matrix."""
        raise NotImplementedError

    @abc.abstractmethod
    def parameters(self) -> np.ndarray:
        """Get underlying representation."""
        raise NotImplementedError

    # Operations.

    @abc.abstractmethod
    def apply(self, target: np.ndarray) -> np.ndarray:
        """Applies group action to a point."""
        raise NotImplementedError

    @abc.abstractmethod
    def multiply(self, other: Self) -> Self:
        """Composes this transformation with another."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def exp(cls, tangent: np.ndarray) -> Self:
        """Computes `expm(wedge(tangent))`."""
        raise NotImplementedError

    @abc.abstractmethod
    def log(self) -> np.ndarray:
        """Computes `vee(logm(transformation matrix))`."""
        raise NotImplementedError

    @abc.abstractmethod
    def adjoint(self) -> np.ndarray:
        """Computes the adjoint."""
        raise NotImplementedError

    @abc.abstractmethod
    def inverse(self) -> Self:
        """Computes the inverse of the transform."""
        raise NotImplementedError

    @abc.abstractmethod
    def normalize(self) -> Self:
        """Normalize/projects values and returns."""
        raise NotImplementedError

    # Plus and minus operators.

    # Eqn. 25.
    def rplus(self, other: np.ndarray) -> Self:
        return self @ self.exp(other)

    # Eqn. 26.
    def rminus(self, other: Self) -> np.ndarray:
        return (other.inverse() @ self).log()

    # Eqn. 27.
    def lplus(self, other: np.ndarray) -> Self:
        return self.exp(other) @ self

    # Eqn. 28.
    def lminus(self, other: Self) -> np.ndarray:
        return (self @ other.inverse()).log()

    def plus(self, other: np.ndarray) -> Self:
        """Alias for rplus."""
        return self.rplus(other)

    def minus(self, other: Self) -> np.ndarray:
        """Alias for rminus."""
        return self.rminus(other)

    # Jacobians.

    @classmethod
    @abc.abstractmethod
    def ljac(cls, other: np.ndarray) -> np.ndarray:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def ljacinv(cls, other: np.ndarray) -> np.ndarray:
        # NOTE: Can just be np.linalg.inv(cls.ljac(other)).
        raise NotImplementedError

    # Eqn. 67.
    @classmethod
    def rjac(cls, other: np.ndarray) -> np.ndarray:
        return cls.ljac(-other)

    @classmethod
    def rjacinv(cls, other: np.ndarray) -> np.ndarray:
        return cls.ljacinv(-other)

    # Eqn. 79.
    def jlog(self) -> np.ndarray:
        return self.rjacinv(self.log())
